######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Contains :class:`IndexingTableModel`.

:author: A. Soininen (VTT)
:date:   25.8.2020
"""
from PySide2.QtCore import QAbstractTableModel, QModelIndex, Qt, Signal
from spinedb_api.spine_io.exporters import gdx


class IndexingTableModel(QAbstractTableModel):
    """
    A table model for parameter_value indexing.

    First column contains the proposed new index keys.
    The rest of the columns contain the parameter values for each set of existing index keys.
    Only selected new index keys are used for indexing.
    Unselected rows are left empty.
    """

    selection_changed = Signal()
    """Emitted after the values have been spread over the selected indexes."""
    manual_selection = Signal()
    """Emitted when the selection has been changed by setData()."""

    def __init__(self, parameter):
        """
        Args:
            parameter (Parameter): a parameter to model
        """
        super().__init__()
        self._records = gdx.LiteralRecords([])
        self._index_name = ""
        self._parameter_values = list(parameter.values)
        self._parameter_nonexpanded_indexes = list(parameter.indexes)
        self._selected = list()
        self._rows_shown = 0
        self._values = [list() for _ in range(len(self._parameter_values))]
        self.records_operational = True

    def get_picking(self):
        """
        Turns the checked record into picking.

        Returns:
            FixedPicking: picked records
        """
        return gdx.FixedPicking(list(self._selected))

    def canFetchMore(self, parent):
        """Returns True if more rows are available to show."""
        return self._rows_shown < len(self._records)

    def clear(self):
        """Clears the model."""
        self.beginResetModel()
        self._records = gdx.LiteralRecords([])
        self._selected = list()
        self._values = [list() for _ in range(len(self._parameter_values))]
        self._rows_shown = 0
        self.endResetModel()
        self.selection_changed.emit()

    def columnCount(self, parent=QModelIndex()):
        """Returns the number of columns."""
        return len(self._parameter_values) + 1

    def data(self, index, role=Qt.DisplayRole):
        """Returns data associated with given model index and role."""
        if role in (Qt.DisplayRole, Qt.ToolTipRole):
            row = index.row()
            column = index.column()
            if column == 0:
                try:
                    value = self._records.records[row][0]
                except gdx.GdxExportException:
                    self.records_operational = False
                    return None
                else:
                    self.records_operational = True
                    return value
            return self._values[column - 1][row]
        if index.column() == 0 and role == Qt.CheckStateRole:
            return Qt.Checked if self._selected[index.row()] else Qt.Unchecked
        return None

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        """Returns header data."""
        if role != Qt.DisplayRole:
            return None
        if orientation == Qt.Vertical:
            return section + 1
        if section == 0:
            return self._index_name
        return ", ".join(self._parameter_nonexpanded_indexes[section - 1])

    def fetchMore(self, parent):
        """Inserts a number of new rows to the table."""
        remainder = len(self._records) - self._rows_shown
        fetch_size = min(remainder, 100)
        end = self._rows_shown + fetch_size
        self.beginInsertRows(parent, self._rows_shown, end - 1)
        for column in self._values:
            column += fetch_size * [None]
        start = self._rows_shown
        self._rows_shown = end
        self._spread_values_over_selected_rows(start)
        self.endInsertRows()

    def flags(self, index):
        """Returns flags for given index."""
        flags = Qt.ItemIsSelectable | Qt.ItemIsEnabled
        if index.column() == 0:
            flags = flags | Qt.ItemIsUserCheckable
        return flags

    def mapped_values_balance(self):
        """
        Returns the balance between available indexes and parameter values.

        Zero means that there is as many indexes available as there are values,
        i.e. the parameter is 'perfectly' indexed.
        A positive value means there are more indexes than values
        while a negative value means there are not enough indexes for all values.

        Returns:
            int: mapped values' balance
        """
        count = sum(1 for selected in self._selected if selected)
        return count - len(self._parameter_values[0].values) if self._parameter_values else 0

    def rowCount(self, parent=QModelIndex()):
        """Return the number of rows."""
        return self._rows_shown

    def select_all(self):
        """Selects all indexes."""
        self._selected = len(self._records) * [True]
        top_left = self.index(0, 0)
        bottom_right = self.index(self._rows_shown - 1, 0)
        self.dataChanged.emit(top_left, bottom_right, [Qt.CheckStateRole])
        self._spread_values_over_selected_rows(0)
        self.selection_changed.emit()

    def setData(self, index, value, role=Qt.EditRole):
        """Sets the checked state for given index."""
        if role != Qt.CheckStateRole:
            return False
        row = index.row()
        self._selected[row] = value == Qt.Checked
        self.dataChanged.emit(index, index, [Qt.CheckStateRole])
        self._spread_values_over_selected_rows(row)
        self.selection_changed.emit()
        self.manual_selection.emit()
        return True

    def set_index_name(self, name):
        """Sets the indexing domain name."""
        self._index_name = name
        self.headerDataChanged.emit(Qt.Horizontal, 0, 0)

    def set_records(self, records, pick_list=None):
        """Overwrites all new indexes."""
        self.beginResetModel()
        self._records = records
        self._selected = pick_list if pick_list is not None else len(records) * [True]
        self._values = [list() for _ in range(len(self._parameter_values))]
        self._rows_shown = 0
        self.selection_changed.emit()
        self.endResetModel()

    def set_picking(self, picking):
        """
        Selects the indexes specified by picking.

        Args:
            picking (Picking): picking
        """
        if picking is None:
            picking = gdx.FixedPicking(len(self._records) * [True])
        pick = picking.pick
        try:
            self._selected = [pick(i) for i in range(len(self._records))]
        except gdx.GdxExportException:
            return
        top_left = self.index(0, 0)
        bottom_right = self.index(self._rows_shown - 1, 0)
        self.dataChanged.emit(top_left, bottom_right, [Qt.CheckStateRole])
        self._spread_values_over_selected_rows(0)
        self.selection_changed.emit()

    def _spread_values_over_selected_rows(self, first_row):
        """Repopulates the table according to selected indexes."""
        value_start = sum(1 for is_selected in self._selected[:first_row] if is_selected)
        for i, parameter_value in enumerate(self._parameter_values):
            value_index = value_start
            value_length = len(parameter_value)
            values = list(parameter_value.values)
            column = self._values[i]
            for j in range(first_row, self._rows_shown):
                is_selected = self._selected[j]
                if is_selected and value_index < value_length:
                    column[j] = str(values[value_index])
                    value_index += 1
                else:
                    column[j] = None
        top_left = self.index(first_row, 1)
        bottom_right = self.index(self._rows_shown - 1, len(self._values))
        self.dataChanged.emit(top_left, bottom_right, [Qt.DisplayRole, Qt.ToolTipRole])



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################
"""
Contains a table view that can accept drops.

:author: A. Soininen (VTT)
:date:   25.5.2021
"""
from PySide2.QtCore import Qt
from PySide2.QtWidgets import QTableView


DROP_MIME_TYPE = "application/spine_items-dtdata"
"""Mime type for drag and drop actions."""


class DropTargetTable(QTableView):
    def dragEnterEvent(self, event):
        super().dragEnterEvent(event)
        if event.mimeData().hasFormat(DROP_MIME_TYPE):
            event.setDropAction(Qt.CopyAction)
            event.acceptProposedAction()



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Contains the SettingsPack class.

:author: A. Soininen (VTT)
:date:   6.5.2020
"""
from spinedb_api.spine_io.exporters import gdx
from .exporter_notifications import ExporterNotifications


class SettingsPack:
    """
    Keeper of all settings and stuff needed for exporting a database.

    Attributes:
        settings (gdx.SetSettings): export settings
        indexing_settings (dict): parameter indexing settings
        merging_settings (dict): parameter merging settings
        none_fallback (NoneFallback): fallback for None parameter values
        none_export (NoneExport): how to handle None values while exporting
    """

    def __init__(self):
        self.settings = None
        self.indexing_settings = None
        self.merging_settings = dict()
        self.none_fallback = gdx.NoneFallback.USE_IT
        self.none_export = gdx.NoneExport.DO_NOT_EXPORT
        self.notifications = ExporterNotifications()

    def to_dict(self):
        """Stores the settings pack into a JSON compatible dictionary."""
        d = dict()
        d["settings"] = self.settings.to_dict() if self.settings is not None else None
        d["indexing_settings"] = (
            gdx.indexing_settings_to_dict(self.indexing_settings) if self.indexing_settings is not None else None
        )
        d["merging_settings"] = {
            parameter_name: [s.to_dict() for s in setting_list]
            for parameter_name, setting_list in self.merging_settings.items()
        }
        d["none_fallback"] = self.none_fallback.value
        d["none_export"] = self.none_export.value
        return d

    @staticmethod
    def from_dict(pack_dict, logger):
        """Restores the settings pack from a dictionary."""
        pack = SettingsPack()
        value = pack_dict.get("none_fallback")
        if value is not None:
            pack.none_fallback = gdx.NoneFallback(value)
        value = pack_dict.get("none_export")
        if value is not None:
            pack.none_export = gdx.NoneExport(value)
        set_settings_dict = pack_dict.get("settings")
        if set_settings_dict is not None:
            try:
                pack.settings = gdx.SetSettings.from_dict(set_settings_dict)
            except gdx.GdxExportException as error:
                logger.msg_error.emit(f"Failed to fully restore GdxExporter settings: {error}")
                return pack
        indexing_settings_dict = pack_dict.get("indexing_settings")
        if indexing_settings_dict is not None:
            try:
                pack.indexing_settings = gdx.indexing_settings_from_dict(indexing_settings_dict)
            except gdx.GdxExportException as error:
                logger.msg_error.emit(f"Failed to fully restore GdxExporter's indexing settings: {error}")
                return pack
        pack.merging_settings = pack_dict.get("merging_settings", dict())
        for name, setting_list in pack.merging_settings.items():
            # For 0.5 compatibility
            if not isinstance(setting_list, list):
                pack.merging_settings[name] = [setting_list]
        pack.merging_settings = {
            parameter_name: [gdx.MergingSetting.from_dict(setting_dict) for setting_dict in setting_list]
            for parameter_name, setting_list in pack.merging_settings.items()
        }
        return pack



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Contains :class:`SetListModel`

:author: A. Soininen (VTT)
:date:   25.8.2020
"""
from PySide2.QtCore import QAbstractListModel, QModelIndex, Qt
from PySide2.QtGui import QColor
from spinedb_api.spine_io.exporters import gdx
from ..list_utils import move_list_elements


class SetListModel(QAbstractListModel):
    """
    A model to configure the domain and set name lists in gdx export settings.

    This model combines domain and set names into a single list.
    The two 'parts' are differentiated by different background colors.
    Items from each part cannot be mixed with the other.
    Both the ordering of the items within each list as well as their exportability flags are handled here.
    """

    def __init__(self, set_settings):
        """
        Args:
            set_settings (gdx.SetSettings): settings whose domain and set name lists should be modelled
        """
        super().__init__()
        self._set_settings = set_settings
        self._sorted_domains = sorted(set_settings.domain_names, key=lambda name: set_settings.domain_tiers[name])
        self._sorted_sets = sorted(set_settings.set_names, key=lambda name: set_settings.set_tiers[name])

    def add_domain(self, domain_name, description, records, origin):
        """
        Adds a new additional domain.

        Args:
            domain_name (str): domain's name
            description (str): domain's description
            records (gdx.Records): domain's sorted records
            origin (gdx.Origin): domain's origin
        """
        exportable = gdx.ExportFlag.EXPORTABLE if origin == gdx.Origin.DATABASE else gdx.ExportFlag.NON_EXPORTABLE
        metadata = gdx.SetMetadata(exportable, origin)
        metadata.description = description
        if not self._set_settings.add_or_replace_domain(domain_name, records, metadata):
            first = len(self._set_settings.domain_names)
            last = first
            self.beginInsertRows(QModelIndex(), first, last)
            self._sorted_domains.append(domain_name)
            self.endInsertRows()

    def drop_domain(self, domain_name):
        """
        Removes a domain.

        Args:
            domain_name (str): name of the domain to remove
        """
        row = self._set_settings.domain_tiers[domain_name]
        self.beginRemoveRows(QModelIndex(), row, row)
        self._set_settings.remove_domain(domain_name)
        self._sorted_domains.pop(row)
        self.endRemoveRows()

    def update_domain(self, domain_name, description, records):
        """
        Updates the records and description of an existing domain.

        Args:
            domain_name (str): domain's name
            description (str): domain's description
            records (gdx.Records): updated records
        """
        self._set_settings.update_records(domain_name, records)
        self._set_settings.metadata(domain_name).description = description
        row = self._set_settings.domain_tiers[domain_name]
        cell = self.index(row, 0)
        self.dataChanged.emit(cell, cell, [Qt.DisplayRole])

    def update_indexing_domains(self, domains, descriptions):
        """
        Updates additional domains needed for parameter index expansion.

        Args:
            domains (dict): a mapping from domain name to records
            descriptions (dict): a mapping from domain name to domain's description
        """
        old_indexing_domain_names = {
            name
            for name in self._set_settings.domain_names
            if self._set_settings.metadata(name).origin == gdx.Origin.INDEXING
        }
        domains_to_drop = old_indexing_domain_names - set(domains.keys())
        for name, records in domains.items():
            description = descriptions[name]
            if name in self._set_settings.domain_names:
                self.update_domain(name, description, records)
            else:
                self.add_domain(name, description, records, gdx.Origin.INDEXING)
        for name in domains_to_drop:
            self.drop_domain(name)

    def data(self, index, role=Qt.DisplayRole):
        """
        Returns the value for given role at given index.

        Qt.DisplayRole returns the name of the domain or set
        while Qt.CheckStateRole returns whether the exportable flag has been set or not.
        Qt.BackgroundRole gives the item's background depending whether it is a domain or a set.

        Args:
            index (QModelIndex): an index to the model
            role (int): the query's role

        Returns:
            the requested value or `None`
        """
        if not index.isValid() or index.column() != 0 or index.row() >= self.rowCount():
            return None
        row = index.row()
        domain_count = len(self._sorted_domains)
        if role == Qt.DisplayRole:
            if row < domain_count:
                return self._sorted_domains[row]
            return self._sorted_sets[row - domain_count]
        if role == Qt.BackgroundRole:
            if row < domain_count:
                if self._set_settings.is_additional(self._sorted_domains[row]):
                    return QColor(235, 235, 110)
                return QColor(245, 245, 120)
            return None
        if role == Qt.CheckStateRole:
            if row < domain_count:
                checked = self._set_settings.is_exportable(self._sorted_domains[row])
            else:
                checked = self._set_settings.is_exportable(self._sorted_sets[row - domain_count])
            return Qt.Checked if checked else Qt.Unchecked
        if role == Qt.ToolTipRole:
            if row < domain_count:
                domain_name = self._sorted_domains[row]
                if domain_name == self._set_settings.global_parameters_domain_name:
                    return "This domain has been set as the global\nparameters domain."
                if self._set_settings.metadata(domain_name).origin == gdx.Origin.INDEXING:
                    return "An additional domain for parameter index expansion."
                if self._set_settings.metadata(domain_name).origin == gdx.Origin.MERGING:
                    return "An additional domain for parameter merging."
        return None

    def flags(self, index):
        """Returns an item's flags."""
        if not index.isValid():
            return Qt.NoItemFlags
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsUserCheckable

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        """Returns an empty string for horizontal header and row number for vertical header."""
        if orientation == Qt.Horizontal:
            return ""
        return section + 1

    def is_domain(self, index):
        """Returns True if index points to a domain name, otherwise returns False."""
        if not index.isValid():
            return False
        return index.row() < len(self._sorted_domains)

    def moveRows(self, sourceParent, sourceRow, count, destinationParent, destinationChild):
        """
        Moves the domain and set names around.

        The names cannot be mixed between domains and sets.

        Args:
            sourceParent (QModelIndex): parent from which the rows are moved
            sourceRow (int): index of the first row to be moved
            count (int): number of rows to move
            destinationParent (QModelIndex): parent to which the rows are moved
            destinationChild (int): index where to insert the moved rows

        Returns:
            True if the operation was successful, False otherwise
        """
        row_count = self.rowCount()
        if destinationChild < 0 or destinationChild >= row_count:
            return False
        last_source_row = sourceRow + count - 1
        domain_count = len(self._sorted_domains)
        # Cannot move domains to ordinary sets and vice versa.
        if sourceRow < domain_count <= last_source_row:
            return False
        if sourceRow < domain_count <= destinationChild:
            return False
        if destinationChild < domain_count <= sourceRow:
            return False
        row_after = destinationChild if sourceRow > destinationChild else destinationChild + 1
        self.beginMoveRows(sourceParent, sourceRow, last_source_row, destinationParent, row_after)
        if sourceRow < domain_count:
            names = self._sorted_domains
        else:
            names = self._sorted_sets
            sourceRow -= domain_count
            last_source_row -= domain_count
            destinationChild -= domain_count
        names[:] = move_list_elements(names, sourceRow, last_source_row, destinationChild)
        tiers = self._set_settings.domain_tiers if sourceRow < domain_count else self._set_settings.set_tiers
        begin = min(sourceRow, destinationChild)
        end = max(sourceRow, destinationChild)
        for row, name in enumerate(names[begin : end + count]):
            tiers[name] = row + begin
        self.endMoveRows()
        return True

    def rowCount(self, parent=QModelIndex()):
        """Returns the number of rows."""
        return len(self._sorted_domains) + len(self._sorted_sets)

    def setData(self, index, value, role=Qt.EditRole):
        """Sets the exportable flag status for given row."""
        if not index.isValid() or role != Qt.CheckStateRole:
            return False
        row = index.row()
        domain_count = len(self._sorted_domains)
        if row < domain_count:
            name = self._sorted_domains[row]
            metadata = self._set_settings.metadata(name)
            exportable = gdx.ExportFlag.EXPORTABLE if value == Qt.Checked else gdx.ExportFlag.NON_EXPORTABLE
            metadata.exportable = exportable
        else:
            metadata = self._set_settings.metadata(self._sorted_sets[row - domain_count])
            exportable = gdx.ExportFlag.EXPORTABLE if value == Qt.Checked else gdx.ExportFlag.NON_EXPORTABLE
            metadata.exportable = exportable
            self.dataChanged.emit(index, index, [Qt.CheckStateRole, Qt.ToolTipRole])
        self.dataChanged.emit(index, index, [Qt.CheckStateRole, Qt.ToolTipRole])
        return True

    def update_global_parameters_domain(self, domain_name):
        previous = self._set_settings.global_parameters_domain_name
        if domain_name == previous:
            return
        self._set_settings.global_parameters_domain_name = domain_name
        if previous:
            row = self._set_settings.domain_tiers[previous]
            index = self.index(row, 0)
            self.dataChanged.emit(index, index, [Qt.CheckStateRole, Qt.ToolTipRole])
        if domain_name:
            row = self._set_settings.domain_tiers[domain_name]
            index = self.index(row, 0)
            self.dataChanged.emit(index, index, [Qt.CheckStateRole, Qt.ToolTipRole])



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Contains utilities shared between project items.

:authors: A. Soininen (VTT)
:date:    1.4.2020
"""

from dataclasses import dataclass
from datetime import datetime
from itertools import dropwhile
import json
import os.path
from pathlib import Path
from time import time
from sqlalchemy import create_engine
from sqlalchemy.engine.url import URL, make_url
import spinedb_api
from spine_engine.utils.queue_logger import SuppressedMessage
from spine_engine.project_item.project_item_resource import file_resource_in_pack, transient_file_resource


EXPORTED_PATHS_FILE_NAME = ".exported.json"
"""Name of the file that stores exporter's internal state."""
EXPORTER_EXECUTION_MANIFEST_FILE_PREFIX = ".export-manifest"
"""Prefix for the temporary files that exporter's executable uses to communicate output paths."""


def database_label(provider_name):
    """Creates a standardized label for database resources.

    Args:
        provider_name (str): resource provider's name

    Returns:
        str: resource label
    """
    return "db_url@" + provider_name


@dataclass
class Database:
    """
    Database specific export settings.
    """

    url: str = ""
    """Database URL."""
    output_file_name: str = ""
    """Output file name; relative to item's data dir."""

    def to_dict(self):
        """
        Serializes :class:`Database` into a dictionary.

        Returns:
            dict: serialized :class:`Database`
        """
        return {"output_file_name": self.output_file_name}

    @staticmethod
    def from_dict(database_dict):
        """
        Deserializes :class:`Database` from a dictionary.

        Args:
            database_dict (dict): serialized :class:`Database`

        Returns:
            Database: deserialized instance
        """
        db = Database()
        db.output_file_name = database_dict["output_file_name"]
        return db


def subdirectory_for_fork(output_file_name, data_dir, output_time_stamps, filter_id_hash):
    """
    Creates scenario/tool based output directory for forked workflow.

    Args:
        output_file_name (str): file name
        data_dir (str): project item's data directory
        output_time_stamps (bool): True if time stamp data should be included in the output path
        filter_id_hash (str): hashed filter id

    Returns:
        str: absolute output path
    """
    if output_time_stamps:
        stamp = datetime.fromtimestamp(time())
        time_stamp = "run@" + stamp.isoformat(timespec="seconds").replace(":", ".")
    else:
        time_stamp = ""
    if filter_id_hash:
        if time_stamp:
            path = os.path.join(data_dir, filter_id_hash + "_" + time_stamp, output_file_name)
        else:
            path = os.path.join(data_dir, filter_id_hash, output_file_name)
    else:
        path = os.path.join(data_dir, time_stamp, output_file_name)
    return path


def exported_files_as_resources(item_name, exported_files, data_dir, databases):
    """Collects exported files from 'export manifests'.

    Args:
        item_name (str): item's name
        exported_files (dict, optional): item's exported files cache
        data_dir (str): item's data directory
        databases (Iterable of Database): item's upstream databases

    Returns:
        tuple: output resources and updated exported files cache
    """
    manifests = collect_execution_manifests(data_dir)
    if manifests is not None:
        names = {db.output_file_name for db in databases}
        manifests = {
            out_file_name: files
            for out_file_name, files in collect_execution_manifests(data_dir).items()
            if out_file_name in names
        }
        exported_files = {name: [str(Path(data_dir, f)) for f in files] for name, files in manifests.items()}
    resources = list()
    if exported_files is not None:
        for db in databases:
            if db.output_file_name:
                files = {f for f in exported_files.get(db.output_file_name, []) if Path(f).exists()}
                if files:
                    resources = [file_resource_in_pack(item_name, db.output_file_name, f) for f in files]
                else:
                    resources.append(transient_file_resource(item_name, db.output_file_name))
    else:
        for db in databases:
            if db.output_file_name:
                resources.append(transient_file_resource(item_name, db.output_file_name))
    return resources, exported_files


def collect_execution_manifests(data_dir):
    """Collects output file names from export manifest files written by exporter's executable item.

    Args:
        data_dir (str): item's data directory

    Returns:
        dict: mapping from output label to list of file paths, or None if no manifest files were found
    """
    manifests = None
    for path in Path(data_dir).iterdir():
        if path.name.startswith(EXPORTER_EXECUTION_MANIFEST_FILE_PREFIX) and path.suffix == ".json":
            with open(path) as manifest_file:
                manifest = json.load(manifest_file)
            for out_file_name, paths in manifest.items():
                relative_paths = list()
                for file_path in paths:
                    p = Path(file_path)
                    if p.is_absolute():
                        # Legacy manifests had absolute paths
                        try:
                            relative_paths.append(p.relative_to(data_dir))
                        except ValueError:
                            # Project may have been moved to another directory (or system)
                            # so data_dir is differs from manifest file content.
                            # Try resolving the relative path manually.
                            parts = tuple(dropwhile(lambda part: part != "output", p.parts))
                            relative_paths.append(str(Path(*parts)))
                    else:
                        relative_paths.append(file_path)
                if manifests is None:
                    manifests = dict()
                manifests.setdefault(out_file_name, list()).extend(paths)
    return manifests


def convert_to_sqlalchemy_url(urllib_url, item_name="", logger=None):
    """Returns a sqlalchemy url from the url or None if not valid."""
    selections = f"<b>{item_name}</b> selections" if item_name else "selections"
    if logger is None:
        logger = _NoLogger()
    if not urllib_url:
        logger.msg_error.emit(f"No URL specified for <b>{item_name}</b>. Please specify one and try again")
        return None
    try:
        url = {key: value for key, value in urllib_url.items() if value}
        dialect = url.pop("dialect")
        if not dialect:
            logger.msg_error.emit(f"Unable to generate URL from {selections}: invalid dialect {dialect}.")
            return None
        if dialect == "sqlite":
            sa_url = URL("sqlite", **url)  # pylint: disable=unexpected-keyword-arg
        else:
            db_api = spinedb_api.SUPPORTED_DIALECTS[dialect]
            drivername = f"{dialect}+{db_api}"
            sa_url = URL(drivername, **url)  # pylint: disable=unexpected-keyword-arg
    except Exception as e:  # pylint: disable=broad-except
        # This is in case one of the keys has invalid format
        logger.msg_error.emit(f"Unable to generate URL from {selections}: {e}")
        return None
    if not sa_url.database:
        logger.msg_error.emit(f"Unable to generate URL from {selections}: database missing.")
        return None
    # Final check
    try:
        engine = create_engine(sa_url)
        with engine.connect():
            pass
    except Exception as e:  # pylint: disable=broad-except
        logger.msg_error.emit(f"Unable to generate URL from {selections}: {e}")
        return None
    return sa_url


class _NoLogger:
    def __getattr__(self, _name):
        return SuppressedMessage()


def split_url_credentials(url):
    sa_url = make_url(url)
    connect_args = sa_url.translate_connect_args()
    username = connect_args.pop("username", None)
    password = connect_args.pop("password", None)
    new_sa_url = URL(sa_url.drivername, **connect_args)
    return str(new_sa_url), (username, password)


def unsplit_url_credentials(url, credentials):
    sa_url = make_url(url)
    connect_args = sa_url.translate_connect_args()
    connect_args["username"], connect_args["password"] = credentials
    new_sa_url = URL(sa_url.drivername, **connect_args)
    return str(new_sa_url)



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Contains base classes for executable items.

:authors: A. Soininen (VTT)
:date:    10.2.2021
"""
import json
import os.path
from pathlib import Path

from spine_engine.project_item.project_item_resource import file_resource_in_pack
from spinedb_api import clear_filter_configs, load_filters
from spinedb_api.filters.tools import filter_configs, name_from_dict
from spinedb_api.spine_io import gdx_utils
from spine_engine.project_item.executable_item_base import ExecutableItemBase
from .utils import EXPORTER_EXECUTION_MANIFEST_FILE_PREFIX


class ExporterExecutableItemBase(ExecutableItemBase):
    """Base class for exporter executable items."""

    def __init__(self, name, output_time_stamps, cancel_on_error, gams_path, project_dir, logger):
        """
        Args:
            name (str): item's name
            output_time_stamps (bool): if True append output directories with time stamps
            cancel_on_error (bool): if True execution fails on all errors else some errors can be ignored
            gams_path (str): GAMS path from Toolbox settings
            project_dir (str): absolute path to project directory
            logger (LoggerInterface): a logger
        """
        super().__init__(name, project_dir, logger)
        self._output_time_stamps = output_time_stamps
        self._cancel_on_error = cancel_on_error
        self._gams_path = gams_path
        self._forks = dict()
        self._result_files = None
        self._process = None

    @staticmethod
    def item_type():
        """See base class."""
        raise NotImplementedError()

    def stop_execution(self):
        """Stops executing this item."""
        super().stop_execution()
        if self._process is not None:
            self._process.terminate()
            self._process = None

    def exclude_execution(self, forward_resources, backward_resources):
        """See base class."""
        manifest_file_name = (
            EXPORTER_EXECUTION_MANIFEST_FILE_PREFIX + (f"-{self.hash_filter_id()}" if self._filter_id else "")
        ) + ".json"
        manifest_file_path = Path(self._data_dir, manifest_file_name)
        if not manifest_file_path.exists():
            return
        with open(Path(self._data_dir, manifest_file_name)) as manifest_file:
            manifest = json.load(manifest_file)
        self._result_files = {label: set(files) for label, files in manifest.items()}

    def _databases_and_forks(self, database_urls):
        """
        Connects output file names to scenario and tool filters.

        Args:
            database_urls (Iterable of str): database URLs.

        Returns:
            tuple: a mapping from full database URL to output file name and
                a mapping from URL to list of scenario/tool/etc. names
        """
        databases = {}
        forks = {}
        for full_url in database_urls:
            url = clear_filter_configs(full_url)
            database = next((db for db in self._databases if db.url == url), None)
            if database is None:
                self._logger.msg_warning.emit(f"<b>{self.name}</b>: No settings for database {url}. Skipping.")
                continue
            if not database.output_file_name:
                self._logger.msg_warning.emit(
                    f"<b>{self.name}</b>: No file name given to export database {url}. Skipping."
                )
                continue
            databases[full_url] = database.output_file_name
            forks[url] = list()
            for config in load_filters(filter_configs(full_url)):
                filter_name = name_from_dict(config)
                if filter_name is not None:
                    forks[url].append(filter_name)
        return databases, forks

    def _output_resources_forward(self):
        """See base class."""
        if self._result_files is None:
            return []
        resources = list()
        for label, output_files in self._result_files.items():
            resources += [file_resource_in_pack(self.name, label, str(Path(self._data_dir, f))) for f in output_files]
        return resources

    def _resolve_gams_system_directory(self):
        """Returns GAMS system path from Toolbox settings or None if GAMS default is to be used.

        Returns:
            str: GAMS system path
        """
        path = self._gams_path
        if not path:
            path = gdx_utils.find_gams_directory()
        if path is not None and os.path.isfile(path):
            path = os.path.dirname(path)
        return path

    @classmethod
    def from_dict(cls, item_dict, name, project_dir, app_settings, specifications, logger):
        """See base class."""
        raise NotImplementedError()



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Undo/redo commands for the View project item.

:authors: M. Marin (ER)
:date:   14.4.2022
"""
from spine_items.commands import SpineToolboxCommand


class PinOrUnpinDBValuesCommand(SpineToolboxCommand):
    """Command to pin or unpin DB values."""

    def __init__(self, view, new_values, old_values):
        """
        Args:
            view (View): the View
            new_values (dict): mapping name to list of value identifiers
            old_values (dict): mapping name to list of value identifiers
        """
        super().__init__()
        self.view = view
        self.new_values = new_values
        self.old_values = old_values
        texts = ["pin" if values else "unpin" + " " + name for name, values in new_values.items()]
        self.setText(f"{', '.join(texts)} in {view.name}")

    def redo(self):
        self.view.do_pin_db_values(self.new_values)

    def undo(self):
        self.view.do_pin_db_values(self.old_values)



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Export item's settings window for .gdx export.

:author: A. Soininen (VTT)
:date:   9.9.2019
"""

from copy import deepcopy
import enum
from PySide2.QtCore import QItemSelection, Qt, Signal, Slot
from PySide2.QtWidgets import QDialogButtonBox, QMessageBox, QWidget
from spinedb_api import SpineDBAPIError
import spinedb_api.spine_io.exporters.gdx as gdx
from ..list_utils import move_selected_elements_by
from ..mvcmodels.record_list_model import RecordListModel
from ..mvcmodels.set_list_model import SetListModel
from .parameter_index_settings_window import ParameterIndexSettingsWindow
from .parameter_merging_settings_window import ParameterMergingSettingsWindow


class State(enum.Enum):
    """Gdx Export Settings window state"""

    OK = enum.auto()
    """Settings are ok."""
    BAD_INDEXING = enum.auto()
    """Not all indexed parameters are set up correctly."""


class GdxExportSettings(QWidget):
    """A setting window for exporting .gdx files."""

    update_requested = Signal(str)
    """Emitted when the Update from database button has been clicked."""
    settings_accepted = Signal()
    """Emitted when the OK button has been clicked."""
    settings_rejected = Signal()
    """Emitted when the Cancel button has been clicked."""

    def __init__(
        self,
        set_settings,
        indexing_settings,
        merging_settings,
        none_fallback,
        none_export,
        url_model,
        exporter_name,
        parent,
    ):
        """
        Args:
            set_settings (gdx.SetSettings): export settings for GAMS sets
            indexing_settings (dict): indexing domain information for indexed parameter values
            merging_settings (dict): parameter merging settings
            none_fallback (NoneFallback): fallback for None parameter values
            none_export (NoneExport): how to handle None values while exporting
            url_model (QAbstractListModel): model that provides database URLs
            exporter_name (str, optional): name of the GdxExporter item that opened the window
            parent (QWidget): a parent widget
        """
        from ..ui.gdx_export_settings import Ui_Form  # pylint: disable=import-outside-toplevel

        super().__init__(parent=parent, f=Qt.Window)
        self._ui = Ui_Form()
        self._ui.setupUi(self)
        self.setWindowTitle(f"Gdx Export settings    -- {exporter_name} --")
        self.setAttribute(Qt.WA_DeleteOnClose, True)
        self._url_model = url_model
        self._ui.database_combo_box.setModel(url_model)
        self._ui.update_button.clicked.connect(self._request_settings_update)
        self._ui.button_box.accepted.connect(self._accept)
        self._ui.button_box.rejected.connect(self._reject)
        self._ui.set_move_up_button.clicked.connect(self._move_sets_up)
        self._ui.set_move_down_button.clicked.connect(self._move_sets_down)
        self._set_settings = set_settings if set_settings is not None else gdx.SetSettings(set(), set(), dict())
        self._populate_global_parameters_combo_box(self._set_settings)
        self._ui.global_parameters_combo_box.currentIndexChanged[str].connect(self._update_global_parameters_domain)
        self._ui.record_sort_alphabetic.clicked.connect(self._sort_records_alphabetically)
        self._ui.record_move_up_button.clicked.connect(self._move_records_up)
        self._ui.record_move_down_button.clicked.connect(self._move_records_down)
        self._set_list_model = SetListModel(self._set_settings)
        self._ui.set_list_view.setModel(self._set_list_model)
        record_list_model = RecordListModel()
        self._ui.record_list_view.setModel(record_list_model)
        self._ui.set_list_view.selectionModel().selectionChanged.connect(self._populate_set_contents)
        self._ui.open_indexed_parameter_settings_button.clicked.connect(self._show_indexed_parameter_settings)
        self._ui.open_parameter_merging_settings_button.clicked.connect(self._show_parameter_merging_settings)
        self._indexing_settings = indexing_settings if indexing_settings is not None else {}
        self._indexed_parameter_settings_window = None
        self._merging_settings = merging_settings
        self._parameter_merging_settings_window = None
        self._none_fallback = none_fallback
        self._none_export = none_export
        self._init_none_fallback_combo_box(none_fallback)
        self._init_none_export_combo_box(none_export)
        self._ui.none_fallback_combo_box.currentTextChanged.connect(self._set_none_fallback)
        self._ui.none_export_combo_box.currentTextChanged.connect(self._set_none_export)
        self._state = State.OK
        self._check_state()

    @property
    def set_settings(self):
        """the settings object"""
        return self._set_settings

    @property
    def indexing_settings(self):
        """indexing settings dict"""
        return self._indexing_settings

    @property
    def merging_settings(self):
        """dictionary of merging settings"""
        return self._merging_settings

    @property
    def none_fallback(self):
        return self._none_fallback

    @property
    def none_export(self):
        return self._none_export

    def reset_settings(self, set_settings, indexing_settings, merging_settings):
        """Resets all settings."""
        if self._indexed_parameter_settings_window is not None:
            self._indexed_parameter_settings_window.close()
            self._indexed_parameter_settings_window = None
        if self._parameter_merging_settings_window is not None:
            self._parameter_merging_settings_window.close()
            self._parameter_merging_settings_window = None
        self._ui.global_parameters_combo_box.clear()
        self._populate_global_parameters_combo_box(set_settings)
        self._set_settings = set_settings
        self._set_list_model = SetListModel(set_settings)
        self._ui.set_list_view.setModel(self._set_list_model)
        self._ui.set_list_view.selectionModel().selectionChanged.connect(self._populate_set_contents)
        self._ui.record_list_view.setModel(RecordListModel())
        self._indexing_settings = indexing_settings
        self._merging_settings = merging_settings
        self._ui.controls_group.setEnabled(True)
        self._ui.button_box.button(QDialogButtonBox.Ok).setEnabled(True)
        self._check_state()

    def settings_reading_cancelled(self):
        self._ui.controls_group.setEnabled(True)
        self._ui.button_box.button(QDialogButtonBox.Ok).setEnabled(True)

    def _check_state(self):
        """Checks if there are parameters in need for indexing."""
        for by_dimensions in self._indexing_settings.values():
            for setting in by_dimensions.values():
                if setting.indexing_domain_name is None:
                    self._ui.indexing_status_label.setText(
                        "<span style='color:#ff3333;white-space: pre-wrap;'>Not all parameters correctly indexed.</span>"
                    )
                    self._state = State.BAD_INDEXING
                    return
        self._state = State.OK
        self._ui.indexing_status_label.setText("")

    @Slot(str)
    def _set_none_fallback(self, option):
        """
        Sets the None fallback option.

        Args:
            option (str): option as a label in the combo box
        """
        if option == "Use it":
            self._none_fallback = gdx.NoneFallback.USE_IT
        else:
            self._none_fallback = gdx.NoneFallback.USE_DEFAULT_VALUE

    def _init_none_fallback_combo_box(self, fallback):
        """
        Sets the current text in None fallback combo box.

        Args:
            fallback (NoneFallback): option
        """
        if fallback == gdx.NoneFallback.USE_IT:
            self._ui.none_fallback_combo_box.setCurrentText("Use it")
        else:
            self._ui.none_fallback_combo_box.setCurrentText("Replace by default value")

    @Slot(str)
    def _set_none_export(self, option):
        """
        Sets the None export option.

        Args:
            option (str): option as a label in the combo box
        """
        if option == "Do not export":
            self._none_export = gdx.NoneExport.DO_NOT_EXPORT
        else:
            self._none_export = gdx.NoneExport.EXPORT_AS_NAN

    def _init_none_export_combo_box(self, export):
        """
        Sets the current text in None export combo box

        Args:
            export (NoneExport): option
        """
        if export == gdx.NoneExport.DO_NOT_EXPORT:
            self._ui.none_export_combo_box.setCurrentText("Do not export")
        else:
            self._ui.none_export_combo_box.setCurrentText("Export as not-a-number")

    def _populate_global_parameters_combo_box(self, settings):
        """(Re)populates the global parameters combo box."""
        self._ui.global_parameters_combo_box.addItem("Nothing selected")
        usable_domains = [name for name in settings.domain_names if not settings.metadata(name).is_additional()]
        self._ui.global_parameters_combo_box.addItems(usable_domains)
        if settings.global_parameters_domain_name:
            self._ui.global_parameters_combo_box.setCurrentText(settings.global_parameters_domain_name)

    def _set_records_ordering_controls_enabled(self, enabled):
        """
        Sets or unsets the enabled state of buttons that control the record key order.

        Args:
            enabled: True if the buttons should be enabled, False otherwise
        """
        self._ui.record_sort_alphabetic.setEnabled(enabled)
        self._ui.record_move_down_button.setEnabled(enabled)
        self._ui.record_move_up_button.setEnabled(enabled)

    @Slot()
    def _accept(self):
        """Emits the settings_accepted signal."""
        if self._state != State.OK:
            QMessageBox.warning(
                self,
                "Bad Parameter Indexing",
                "Parameter indexing not set up correctly. Click 'Indexed parameters...' to open the settings window.",
            )
            return
        self.settings_accepted.emit()
        self.hide()

    @Slot(bool)
    def _move_sets_up(self, checked=False):
        """Moves selected domains and sets up one position."""
        move_selected_elements_by(self._ui.set_list_view, -1)

    @Slot(bool)
    def _move_sets_down(self, checked=False):
        """Moves selected domains and sets down one position."""
        move_selected_elements_by(self._ui.set_list_view, 1)

    @Slot(bool)
    def _move_records_up(self, checked=False):
        """Moves selected records up and position."""
        move_selected_elements_by(self._ui.record_list_view, -1)

    @Slot(bool)
    def _move_records_down(self, checked=False):
        """Moves selected records down on position."""
        move_selected_elements_by(self._ui.record_list_view, 1)

    @Slot()
    def _reject(self):
        """Closes the window."""
        self.close()

    def closeEvent(self, event):
        super().closeEvent(event)
        self.settings_rejected.emit()

    @Slot(bool)
    def _request_settings_update(self, _=True):
        """Requests for fresh settings to be read from the database."""
        url = self._ui.database_combo_box.currentText()
        if not url:
            return
        self._ui.controls_group.setEnabled(False)
        self._ui.button_box.button(QDialogButtonBox.Ok).setEnabled(False)
        self.update_requested.emit(url)

    @Slot(str)
    def _update_global_parameters_domain(self, text):
        """Updates the global parameters domain name."""
        if text == "Nothing selected":
            text = ""
        self._set_list_model.update_global_parameters_domain(text)

    @Slot(QItemSelection, QItemSelection)
    def _populate_set_contents(self, selected, _):
        """Populates the record list by the selected domain's or set's records."""
        selected_indexes = selected.indexes()
        if not selected_indexes:
            return
        selected_set_name = self._set_list_model.data(selected_indexes[0])
        records = self._set_settings.records(selected_set_name)
        record_model = self._ui.record_list_view.model()
        record_model.reset(records, selected_set_name)
        self._set_records_ordering_controls_enabled(records.is_shufflable())

    @Slot(bool)
    def _sort_records_alphabetically(self, _):
        """Sorts the lists of set records alphabetically."""
        model = self._ui.record_list_view.model()
        model.sort_alphabetically()

    @Slot(bool)
    def _show_indexed_parameter_settings(self, _):
        """Shows the indexed parameter settings window."""
        if self._indexed_parameter_settings_window is None:
            url = self._ui.database_combo_box.currentText()
            indexing_settings = deepcopy(self._indexing_settings)
            try:
                self._indexed_parameter_settings_window = ParameterIndexSettingsWindow(
                    indexing_settings, self._set_settings, url, self._none_fallback, self
                )
            except (gdx.GdxExportException, SpineDBAPIError) as error:
                QMessageBox.warning(self, "Failed to open indexing settings", str(error))
                return
            self._indexed_parameter_settings_window.settings_approved.connect(self._gather_parameter_indexing_settings)
            self._indexed_parameter_settings_window.settings_rejected.connect(
                self._dispose_parameter_indexing_settings_window
            )
        self._indexed_parameter_settings_window.show()

    @Slot(bool)
    def _show_parameter_merging_settings(self, _):
        """Shows the parameter merging settings window."""
        if self._parameter_merging_settings_window is None:
            url = self._ui.database_combo_box.currentText()
            self._parameter_merging_settings_window = ParameterMergingSettingsWindow(
                self._merging_settings, self._set_settings, url, self
            )
            self._parameter_merging_settings_window.settings_approved.connect(self._parameter_merging_approved)
            self._parameter_merging_settings_window.settings_rejected.connect(self._dispose_parameter_merging_window)
        self._parameter_merging_settings_window.show()

    @Slot()
    def _gather_parameter_indexing_settings(self):
        """Gathers settings from the indexed parameters settings window."""
        self._indexing_settings = self._indexed_parameter_settings_window.indexing_settings
        indexing_domains, descriptions = self._indexed_parameter_settings_window.additional_indexing_domains()
        self._set_list_model.update_indexing_domains(indexing_domains, descriptions)
        self._state = State.OK
        self._ui.indexing_status_label.setText("")

    @Slot()
    def _parameter_merging_approved(self):
        """Collects merging settings from the parameter merging window."""
        new_merging_settings = self._parameter_merging_settings_window.merging_settings
        old_domain_names = {s.new_domain_name for setting_list in self._merging_settings.values() for s in setting_list}
        new_domain_names = {s.new_domain_name for setting_list in new_merging_settings.values() for s in setting_list}
        new_records = dict()
        new_descriptions = dict()
        for setting_list in new_merging_settings.values():
            for setting in setting_list:
                merge_records = gdx.merging_records(setting)
                records = new_records.get(setting.new_domain_name)
                if records is None:
                    new_records[setting.new_domain_name] = merge_records
                else:
                    combined_records = gdx.LiteralRecords(records.records + merge_records.records)
                    new_records[setting.new_domain_name] = combined_records
                new_descriptions[setting.new_domain_name] = setting.new_domain_description
        for domain_to_drop in old_domain_names - new_domain_names:
            self._set_list_model.drop_domain(domain_to_drop)
        for domain_to_update in old_domain_names & new_domain_names:
            self._set_list_model.update_domain(
                domain_to_update, new_descriptions[domain_to_update], new_records[domain_to_update]
            )
        for domain_to_add in new_domain_names - old_domain_names:
            self._set_list_model.add_domain(
                domain_to_add, new_descriptions[domain_to_add], new_records[domain_to_add], gdx.Origin.MERGING
            )
        self._merging_settings = new_merging_settings

    @Slot()
    def _dispose_parameter_indexing_settings_window(self):
        """Removes references to the indexed parameter settings window."""
        self._indexed_parameter_settings_window = None

    @Slot()
    def _dispose_parameter_merging_window(self):
        """Removes references to the parameter merging settings window."""
        self._parameter_merging_settings_window = None



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Merger's execute kernel (do_work), as target for a multiprocess.Process

:authors: M. Marin (KTH)
:date:   6.11.2020
"""
import os
from spine_engine.utils.helpers import create_log_file_timestamp, remove_credentials_from_url
from spinedb_api import (
    clear_filter_configs,
    export_data,
    import_data,
    SpineDBAPIError,
    SpineDBVersionError,
    DatabaseMapping,
    create_new_spine_database,
)


def _get_db_map(url, logger, purge_before_writing=False):
    if purge_before_writing:
        create_new_spine_database(url)
    try:
        db_map = DatabaseMapping(url)
    except (SpineDBAPIError, SpineDBVersionError) as err:
        sanitized_url = remove_credentials_from_url(url)
        logger.msg_warning.emit(f"Skipping url <b>{clear_filter_configs(sanitized_url)}</b>: {err}")
        return None
    return db_map


def do_work(cancel_on_error, purge_before_writing, logs_dir, from_urls, to_urls, logger):
    from_db_maps_iter = (_get_db_map(url, logger) for url in from_urls)
    to_db_maps_iter = (_get_db_map(url, logger, purge_before_writing) for url in to_urls)
    to_db_maps = [db_map for db_map in to_db_maps_iter if db_map is not None]
    from_db_map_data = {db_map: export_data(db_map) for db_map in from_db_maps_iter if db_map is not None}
    all_errors = []
    for from_db_map, data in from_db_map_data.items():
        for to_db_map in to_db_maps:
            import_count, import_errors = import_data(to_db_map, **data)
            all_errors += import_errors
            if import_errors and cancel_on_error and to_db_map.has_pending_changes():
                to_db_map.rollback_session()
                continue
            sanitized_from_url = remove_credentials_from_url(from_db_map.db_url)
            sanitized_to_url = remove_credentials_from_url(to_db_map.db_url)
            if import_count:
                to_db_map.commit_session(f"Import {import_count} items from {sanitized_from_url}")
                logger.msg_success.emit(
                    "Merged {0} items with {1} errors from {2} into {3}".format(
                        import_count, len(import_errors), sanitized_from_url, sanitized_to_url
                    )
                )
            else:
                logger.msg_warning.emit(
                    "No new data merged from {0} into {1}".format(sanitized_from_url, sanitized_to_url)
                )
    for db_map in from_db_map_data:
        db_map.connection.close()
    for db_map in to_db_maps:
        db_map.connection.close()
    if all_errors:
        # Log errors in a time stamped file into the logs directory
        timestamp = create_log_file_timestamp()
        logfilepath = os.path.abspath(os.path.join(logs_dir, timestamp + "_error.log"))
        with open(logfilepath, "w") as f:
            for err in all_errors:
                f.write("{0}\n".format(err))
        # Make error log file anchor with path as tooltip
        logfile_anchor = f"<a style='color:#BB99FF;' title='{logfilepath}' href='file:///{logfilepath}'>error log</a>"
        logger.msg_error.emit("Import errors. Logfile: {0}".format(logfile_anchor))
    return (bool(from_db_map_data),)



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Undo/redo commands for the GdxExporter project item.

:authors: A. Soininen (VTT)
:date:   30.4.2020
"""
from spine_items.commands import SpineToolboxCommand


class UpdateSettings(SpineToolboxCommand):
    """Command to update GdxExporter settings."""

    def __init__(self, exporter, settings, indexing_settings, merging_settings, none_fallback, none_export):
        """
        Args:
            exporter (GdxExporter): the GdxExporter
            settings (SetSettings): gdx settings
            indexing_settings (dict): parameter indexing settings
            merging_settings (dict): parameter merging settings
            none_fallback (NoneFallback): fallback option on None values
            none_export (NoneExport): how to handle Nones while exporting
        """
        super().__init__()
        self._exporter = exporter
        self._redo_settings_tuple = (settings, indexing_settings, merging_settings, none_fallback, none_export)
        p = exporter.settings_pack()
        self._undo_settings_tuple = (p.settings, p.indexing_settings, p.merging_settings)
        self.setText(f"change settings of {exporter.name}")

    def redo(self):
        self._exporter.undo_or_redo_settings(*self._redo_settings_tuple)

    def undo(self):
        self._exporter.undo_or_redo_settings(*self._undo_settings_tuple)



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################
"""
Contains :class:`OutputChannel` class.

:authors: A. Soininen (VTT)
:date:    4.1.2022
"""
from dataclasses import dataclass


@dataclass
class OutputChannel:
    """
    Input resource specific export settings.
    """

    in_label: str
    """Label of input resource."""
    out_label: str = ""
    """Label of output resource. Output file name in case of single file export."""

    def to_dict(self):
        """
        Serializes :class:`OutputChannel` into a dictionary.

        Returns:
            dict: serialized :class:`OutputChannel`
        """
        return {"in_label": self.in_label, "out_label": self.out_label}

    @staticmethod
    def from_dict(channel_dict):
        """
        Deserializes :class:`OutputChannel` from a dictionary.

        Args:
            channel_dict (dict): serialized :class:`OutputChannel`

        Returns:
            OutputChannel: deserialized instance
        """
        channel = OutputChannel(channel_dict["in_label"], channel_dict["out_label"])
        return channel



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Undo/redo commands for the DataStore project item.

:authors: M. Marin (KTH)
:date:   5.5.2020
"""
from spine_items.commands import SpineToolboxCommand


class UpdateDSURLCommand(SpineToolboxCommand):
    def __init__(self, ds, **kwargs):
        """Command to update DS url.

        Args:
            ds (DataStore): the DS
            kwargs: url keys and their values
        """
        super().__init__()
        self.ds = ds
        self.redo_kwargs = kwargs
        self.undo_kwargs = {k: self.ds._url[k] for k in kwargs}
        if len(kwargs) == 1:
            self.setText(f"change {list(kwargs.keys())[0]} of {ds.name}")
        else:
            self.setText(f"change url of {ds.name}")

    def redo(self):
        self.ds.do_update_url(**self.redo_kwargs)

    def undo(self):
        self.ds.do_update_url(**self.undo_kwargs)



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################
"""
Contains :class:`DomainNameListModel`

:author: A. Soininen (VTT)
:date:   17.9.2020
"""
from PySide2.QtCore import QAbstractListModel, QModelIndex, Qt


class DomainNameListModel(QAbstractListModel):
    """
    Model for domains_list_view.

    Stores EntityClassInfo objects displaying the entity name in domains_list_view.
    """

    def __init__(self, entity_classes):
        """
        Args:
            entity_classes (list): a list of EntityClassObjects
        """
        super().__init__()
        self._entity_classes = entity_classes

    def data(self, index, role=Qt.DisplayRole):
        """Returns model's data for given index."""
        if role != Qt.DisplayRole or not index.isValid():
            return None
        return self._entity_classes[index.row()].name

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        """Returns None."""
        return None

    def index_for(self, set_name):
        """Returns the QModelIndex for given set name."""
        try:
            row = [entity_class.name for entity_class in self._entity_classes].index(set_name)
        except ValueError:
            return QModelIndex()
        else:
            return self.index(row, 0)

    def item_at(self, row):
        """Returns the EntityClassInfo object at given row."""
        return self._entity_classes[row]

    def rowCount(self, parent=QModelIndex()):
        """Returns the size of the model."""
        return len(self._entity_classes)



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
A worker based machinery to construct the settings data structures needed for gdx export outside the UI loop.

:author: A. Soininen (VTT)
:date:   19.12.2019
"""

from copy import deepcopy
from PySide2.QtCore import QObject, QThread, Signal, Slot
from spinedb_api import DatabaseMapping, SpineDBAPIError
from spinedb_api.spine_io.exporters import gdx


class Worker(QObject):
    """
    A worker to construct export settings for a database.

    Attributes:
        thread (QThread): the thread the worker executes in
        database_url (str): database URL
    """

    database_unavailable = Signal()
    """Emitted when opening the database fails."""
    errored = Signal(object)
    """Emitted when an error occurs."""
    finished = Signal(object)
    """Emitted when the worker has finished."""

    def __init__(self, database_url, none_fallback, logger):
        """
        Args:
            database_url (str): database's URL
            none_fallback (NoneFallback): how to handle None parameter values
            logger (LoggerInterface): a logger
        """
        super().__init__()
        self.thread = QThread()
        self.moveToThread(self.thread)
        self._none_fallback = none_fallback
        self.database_url = str(database_url)
        self._previous_settings = None
        self._previous_indexing_settings = None
        self._previous_merging_settings = None
        self._logger = logger
        self.thread.started.connect(self._fetch_settings)

    @Slot()
    def _fetch_settings(self):
        """Constructs settings and parameter index settings."""
        result = _Result(*self._read_settings())
        if result.set_settings is None:
            return
        if self._previous_settings is not None:
            updated_settings = deepcopy(self._previous_settings)
            try:
                updated_settings.update(result.set_settings)
            except gdx.GdxExportException as error:
                self.errored.emit(str(error))
                return
            updated_indexing_settings = self._update_indexing_settings(result.indexing_settings)
            if updated_indexing_settings is None:
                return
            updated_merging_settings = self._update_merging_settings(updated_settings)
            if updated_merging_settings is None:
                return
            result.set_settings = updated_settings
            result.indexing_settings = updated_indexing_settings
            result.merging_settings = updated_merging_settings
        self.finished.emit(result)
        self.thread.quit()

    def set_previous_settings(self, previous_settings, previous_indexing_settings, previous_merging_settings):
        """
        Makes worker update existing settings instead of just making new ones.

        Args:
            previous_settings (gdx.SetSettings): existing set settings
            previous_indexing_settings (dict): existing indexing settings
            previous_merging_settings (dict): existing merging settings
        """
        self._previous_settings = previous_settings
        self._previous_indexing_settings = previous_indexing_settings
        self._previous_merging_settings = previous_merging_settings

    def _read_settings(self):
        """Reads fresh gdx settings from the database."""
        try:
            database_map = DatabaseMapping(self.database_url)
        except SpineDBAPIError:
            self.database_unavailable.emit()
            return None, None
        try:
            settings = gdx.make_set_settings(database_map)
            indexing_settings = gdx.make_indexing_settings(database_map, self._none_fallback, self._logger)
        except gdx.GdxExportException as error:
            self.errored.emit(error)
            return None, None
        finally:
            database_map.connection.close()
        return settings, indexing_settings

    def _update_indexing_settings(self, new_indexing_settings):
        """Updates the parameter indexing settings according to changes in the database."""
        updated_indexing_settings = gdx.update_indexing_settings(
            self._previous_indexing_settings, new_indexing_settings
        )
        return updated_indexing_settings

    def _update_merging_settings(self, updated_settings):
        """Updates the parameter merging settings according to changes in the database"""
        try:
            database_map = DatabaseMapping(self.database_url)
        except SpineDBAPIError as error:
            self.errored.emit(error)
            return None
        try:
            updated_merging_settings = gdx.update_merging_settings(
                self._previous_merging_settings, updated_settings, database_map
            )
        except gdx.GdxExportException as error:
            self.errored.emit(error)
            return None
        finally:
            database_map.connection.close()
        return updated_merging_settings


class _Result:
    """
    Contains fetched export settings.

    Attributes:
        set_settings (gdx.SetSettings): gdx export settings
        indexing_settings (dict): parameter indexing settings
        merging_settings (dict): parameter merging settings
    """

    def __init__(self, set_settings, indexing_settings):
        """
        Args:
            set_settings (gdx.SetSettings): gdx export settings
            indexing_settings (dict): parameter indexing settings
        """
        self.set_settings = set_settings
        self.indexing_settings = indexing_settings
        self.merging_settings = dict()



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Contains GdxExporter's executable item as well as support utilities.

:authors: A. Soininen (VTT)
:date:    2.4.2020
"""
from json import dump
from pathlib import Path
from spinedb_api import clear_filter_configs
from spinedb_api.spine_io.exporters import gdx
from spine_engine.utils.serialization import deserialize_path
from spine_engine.utils.returning_process import ReturningProcess
from spine_engine.spine_engine import ItemExecutionFinishState
from spine_items.utils import Database, EXPORTER_EXECUTION_MANIFEST_FILE_PREFIX
from ..exporter_executable_item_base import ExporterExecutableItemBase
from .do_work import do_work
from .item_info import ItemInfo
from .settings_pack import SettingsPack


class ExecutableItem(ExporterExecutableItemBase):
    def __init__(
        self, name, settings_pack, databases, output_time_stamps, cancel_on_error, gams_path, project_dir, logger
    ):
        """
        Args:
            name (str): item's name
            settings_pack (SettingsPack): export settings
            databases (list of Database): database export settings
            output_time_stamps (bool): if True append output directories with time stamps
            cancel_on_error (bool): if True execution fails on all errors else some errors can be ignored
            gams_path (str): GAMS path from Toolbox settings
            project_dir (str): absolute path to project directory
            logger (LoggerInterface): a logger
        """
        super().__init__(name, output_time_stamps, cancel_on_error, gams_path, project_dir, logger)
        self._databases = databases
        self._settings_pack = settings_pack

    @staticmethod
    def item_type():
        """Returns GdxExporter's type identifier string."""
        return ItemInfo.item_type()

    def _database_output_file_names(self, database_urls):
        """
        Connects full database urls to output file names.

        Args:
            database_urls (Iterable of str): full database URLs including filters

        Returns:
            dict: a mapping from full database URL to output file name
        """
        databases = {}
        for full_url in database_urls:
            url = clear_filter_configs(full_url)
            database = next((db for db in self._databases if db.url == url), None)
            if database is None:
                self._logger.msg_warning.emit(f"<b>{self.name}</b>: No settings for database {url}. Skipping.")
                continue
            if not database.output_file_name:
                self._logger.msg_warning.emit(
                    f"<b>{self.name}</b>: No file name given to export database {url}. Skipping."
                )
                continue
            databases[full_url] = database.output_file_name
        return databases

    def execute(self, forward_resources, backward_resources):
        """See base class."""
        if not super().execute(forward_resources, backward_resources):
            return ItemExecutionFinishState.FAILURE
        if self._settings_pack.settings is None:
            self._logger.msg_warning.emit(f"<b>{self.name}</b>: No export settings configured. Skipping.")
            return ItemExecutionFinishState.SKIPPED
        database_urls = [r.url for r in forward_resources if r.type_ == "database"]
        gams_system_directory = self._resolve_gams_system_directory()
        if gams_system_directory is None:
            self._logger.msg_error.emit(f"<b>{self.name}</b>: Cannot proceed. No GAMS installation found.")
            return ItemExecutionFinishState.FAILURE
        databases, self._forks = self._databases_and_forks(database_urls)
        self._process = ReturningProcess(
            target=do_work,
            args=(
                self._settings_pack,
                self._output_time_stamps,
                self._cancel_on_error,
                self._data_dir,
                gams_system_directory,
                databases,
                self._forks,
                self._logger,
            ),
        )
        result = self._process.run_until_complete()
        # result contains only the success flag if execution was forcibly stopped.
        if len(result) > 1:
            self._result_files = result[1]
            file_name = (
                EXPORTER_EXECUTION_MANIFEST_FILE_PREFIX + (f"-{self.hash_filter_id()}" if self._filter_id else "")
            ) + ".json"
            with open(Path(self._data_dir, file_name), "w") as manifest:
                dump(
                    {
                        url: list(str(Path(file).relative_to(Path(self._data_dir))) for file in files)
                        for url, files in self._result_files.items()
                    },
                    manifest,
                )
        self._process = None
        return ItemExecutionFinishState.SUCCESS if result[0] else ItemExecutionFinishState.FAILURE

    @classmethod
    def from_dict(cls, item_dict, name, project_dir, app_settings, specifications, logger):
        """See base class."""
        try:
            settings_pack = SettingsPack.from_dict(item_dict["settings_pack"], logger)
        except gdx.GdxExportException as error:
            logger.msg_error.emit(f"Failed to fully restore GdxExporter settings: {error}")
            settings_pack = SettingsPack()
        databases = list()
        for db_dict in item_dict["databases"]:
            db = Database.from_dict(db_dict)
            db.url = deserialize_path(db_dict["database_url"], project_dir)
            databases.append(db)
        output_time_stamps = item_dict.get("output_time_stamps", False)
        cancel_on_error = item_dict.get("cancel_on_error", True)
        gams_path = app_settings.value("appSettings/gamsPath", defaultValue=None)
        return cls(name, settings_pack, databases, output_time_stamps, cancel_on_error, gams_path, project_dir, logger)



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Contains :class:`IndexingDomainListModel`.

:author: A. Soininen (VTT)
:date:   25.8.2020
"""
from PySide2.QtCore import QAbstractListModel, QModelIndex, Qt, Signal
from spinedb_api.spine_io.exporters import gdx


class IndexingDomainListItem:
    """
    Holds additional indexing domain information.

    Attributes:
        name (str): domain's name
        description (str): domain's description
        expression (str or NoneType): record key generator expression, or None
        length (int): length of the domain
        extract_from_parameter_name (str): parameter from which record keys are to be extracted
        extract_from_parameter_domain_names (tuple): parameter's domain names from which record keys are to be extracted
    """

    def __init__(self, name):
        """
        Args:
            name (str): domain's name
        """
        self.name = name
        self.description = ""
        self.expression = None
        self.length = 0
        self.extract_from_parameter_name = None
        self.extract_from_parameter_domain_names = None


class IndexingDomainListModel(QAbstractListModel):
    """A model to manage additional domains needed for indexed parameter expansion."""

    domains_added = Signal(list)
    """Emitted after new domains have been added to the model."""
    domains_removed = Signal(list)
    """Emitted after domains have been removed from this model"""
    domain_renamed = Signal(str, str)
    """Emitted after a domain has been renamed."""
    indexes_changed = Signal(str, object)
    """Emitted when a domain's records change."""

    def __init__(self, set_settings, parameters):
        """
        Args:
            set_settings (SetSettings): export settings
            parameters (dict): indexed parameters
        """
        super().__init__()
        self._set_settings = set_settings
        self._domains = list()
        for name in set_settings.domain_names:
            metadata = set_settings.metadata(name)
            if metadata.origin != gdx.Origin.INDEXING:
                continue
            item = IndexingDomainListItem(name)
            item.description = metadata.description
            records = set_settings.records(name)
            if isinstance(records, gdx.GeneratedRecords):
                item.expression = records.expression
                item.length = len(records)
            elif isinstance(records, gdx.ExtractedRecords):
                item.extract_from_parameter_name = records.parameter_name
                if records.domain_names is None:
                    item.extract_from_parameter_domain_names = next(
                        iter(parameters[records.parameter_name].values())
                    ).domain_names
                else:
                    item.extract_from_parameter_domain_names = records.domain_names
            self._domains.append(item)

    def create_new_domain(self):
        """Adds a new domain as the last element in the list."""
        self.insertRows(self.rowCount(), 1)

    def data(self, index, role=Qt.DisplayRole):
        """
        Returns the domain name at given row on ``DisplayRole``.

        Args:
            index (QModelIndex): an index to the list
            role (int): data role
        """
        if role == Qt.DisplayRole:
            return self._domains[index.row()].name
        return None

    def item_at(self, row):
        """
        Returns :class:`IndexingDomainListItem` at the given row.

        Args:
            row (int): a row in the list.

        Returns:
            IndexingDomainListItem: item at the given row
        """
        return self._domains[row]

    def flags(self, index):
        """
        Returns item flags.

        Args:
            index (QModelIndex): list index

        Returns:
            int: item flags
        """
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsEditable

    def gather_domains(self, parameters):
        """
        Returns domain name and records.

        Args:
            parameters (dict): parameters

        Returns:
            tuple: mappings from domain name to records and from domain name to description
        """
        domains = dict()
        descriptions = dict()
        for item in self._domains:
            descriptions[item.name] = item.description
            if item.expression is not None:
                domains[item.name] = gdx.GeneratedRecords(item.expression, item.length)
            elif item.extract_from_parameter_name is not None:
                parameter = parameters[item.extract_from_parameter_name][item.extract_from_parameter_domain_names]
                value_indexes = [(str(i),) for i in next(iter(parameter.values)).indexes]
                domains[item.name] = gdx.ExtractedRecords(
                    item.extract_from_parameter_name, item.extract_from_parameter_domain_names, value_indexes
                )
            else:
                domains[item.name] = gdx.LiteralRecords([])
        return domains, descriptions

    def insertRows(self, row, count, parent=QModelIndex()):
        """
        Inserts new rows to the list.

        Args:
            row (int): first row occupied by the inserted items
            count (int): number of inserted items
            parent (QModelIndex): ignored

        Returns:
            bool: True if the operation was successful
        """
        self.beginInsertRows(parent, row, row + count - 1)
        reserved = {d.name for d in self._domains}
        names = list()
        number = 1
        while len(names) != count:
            name = f"t{number}"
            if name not in reserved:
                reserved.add(name)
                names.append(name)
            number += 1
        new_domains = [IndexingDomainListItem(name) for name in names]
        self._domains = self._domains[:row] + new_domains + self._domains[row:]
        self.endInsertRows()
        self.domains_added.emit(names)
        return True

    def removeRows(self, row, count, parent=QModelIndex()):
        """
        Removes rows.

        Args:
            row (int): first row to remove
            count (int): number of rows to remove
            parent (QModelIndex): ignored

        Returns:
            bool: True if the operations was successful
        """
        self.beginRemoveRows(parent, row, row + count - 1)
        removed = self._domains[row : row + count]
        self._domains = self._domains[:row] + self._domains[row + count :]
        self.endRemoveRows()
        self.domains_removed.emit([item.name for item in removed])
        return True

    def remove_rows(self, rows):
        """
        Removes non-contiguous set of rows effectively resetting the model.

        Args:
            rows (list of int): rows to remove
        """
        if len(rows) == 1:
            self.removeRows(rows[0], 1)
            return
        if all(previous_row + 1 == next_row for previous_row, next_row in zip(rows[:-1], rows[1:])):
            self.removeRows(rows[0], len(rows))
            return
        self.beginResetModel()
        removed = [d for i, d in enumerate(self._domains) if i in rows]
        self._domains = [d for i, d in enumerate(self._domains) if i not in rows]
        self.endResetModel()
        self.domains_removed.emit([item.name for item in removed])

    def rowCount(self, parent=QModelIndex()):
        """
        Returns the list length.

        Returns:
            int: number of rows in the list
        """
        return len(self._domains)

    def setData(self, index, value, role=Qt.EditRole):
        """
        Sets domain name at given index on ``EditRole``.

        Args:
            index (QModelIndex): an index to the list
            value (str): new domain name
            role (int): role

        Returns:
            bool: True if a domain name was changed
        """
        if role != Qt.EditRole:
            return False
        reserved = {d.name for d in self._domains} | self._set_settings.set_names
        if value in reserved:
            return False
        item = self._domains[index.row()]
        old_name = item.name
        item.name = value
        self.domain_renamed.emit(old_name, value)
        self.dataChanged.emit(index, index, [Qt.DisplayRole])
        return True



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################
"""
Contains :class:`ParameterNameListModel`

:author: A. Soininen (VTT)
:date:   17.9.2020
"""
from PySide2.QtCore import QAbstractListModel, QModelIndex, Qt


class ParameterNameListModel(QAbstractListModel):
    """Model for parameter_name_list_view."""

    def __init__(self, names):
        """
        Args:
            names (list): list of parameter names to show in the view
        """
        super().__init__()
        self._names = names
        self._selected = len(names) * [True]
        self._selected_cache = dict()
        self._current_entity_class = None

    def data(self, index, role=Qt.DisplayRole):
        """Returns the model's data."""
        if not index.isValid():
            return None
        if role == Qt.DisplayRole:
            return self._names[index.row()]
        if role == Qt.CheckStateRole:
            return Qt.Checked if self._selected[index.row()] else Qt.Unchecked
        return None

    def flags(self, index):
        """Returns flags for given index."""
        if not index.isValid():
            return Qt.NoItemFlags
        return Qt.ItemIsSelectable | Qt.ItemIsUserCheckable | Qt.ItemIsEnabled

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        """Returns None."""
        return None

    def known_selections(self, entity_class_name, selected):
        """
        Stores known selected parameter names for later use.

        Args:
            entity_class_name (str): domain names that identify the selected parameter names
            selected (list of str): list of selected parameter names
        """
        self._selected_cache[entity_class_name] = selected

    def reset_names(self, names, entity_class_name):
        """
        Resets the model's parameter names when a new index is selected in domains_list_view.

        Args:
            names (list of str): new parameter names
            entity_class_name (str): parameters'
        """
        self.beginResetModel()
        if self._current_entity_class is not None:
            self._selected_cache[self._current_entity_class] = [
                name for name, selected in zip(self._names, self._selected) if selected
            ]
        self._names = names
        selected = self._selected_cache.get(entity_class_name)
        if selected is not None:
            self._selected = len(names) * [False]
            name_indexes = {name: i for i, name in enumerate(self._names)}
            for name in selected:
                self._selected[name_indexes[name]] = True
        else:
            self._selected = len(names) * [True]
        self._current_entity_class = entity_class_name
        self.endResetModel()

    def rowCount(self, parent=QModelIndex()):
        """Returns the number of parameter names."""
        return len(self._names)

    def select(self, names):
        """Selects parameters for inclusion in the merged parameter."""
        for i, existing_name in enumerate(self._names):
            self._selected[i] = existing_name in names

    def selected(self):
        """Returns a list of the selected parameters."""
        return [name for name, select in zip(self._names, self._selected) if select]

    def setData(self, index, value, role=Qt.EditRole):
        """Selects or deselects the parameter at given index for inclusion in the merged parameter."""
        if role != Qt.CheckStateRole or not index.isValid():
            return False
        self._selected[index.row()] = value == Qt.Checked
        return True



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Provides an optional widget for Tool Specification Editor for each Tool Spec type (julia, python, executable, gams).

:author: P. Savolainen (VTT)
:date:   12.2.2021
"""

from PySide2.QtCore import Slot
from PySide2.QtWidgets import QWidget
from PySide2.QtGui import QStandardItemModel, QStandardItem
from spine_engine.utils.helpers import resolve_conda_executable, resolve_python_interpreter
from spine_engine.execution_managers.conda_kernel_spec_manager import CondaKernelSpecManager
from spinetoolbox.helpers import busy_effect, file_is_valid, select_python_interpreter
from spinetoolbox.widgets.notification import Notification
from spinetoolbox.widgets.kernel_editor import KernelEditor, find_python_kernels


class OptionalWidget(QWidget):
    def __init__(self, parent):
        super().__init__()
        self._parent = parent

    @property
    def _toolbox(self):
        return self._parent._toolbox

    @property
    def _project(self):
        return self._toolbox._project

    @property
    def _settings(self):
        return self._project._settings

    @property
    def _logger(self):
        return self._toolbox._logger

    def init_widget(self, specification):
        raise NotImplementedError

    def add_execution_settings(self):
        raise NotImplementedError


class PythonToolSpecOptionalWidget(OptionalWidget):
    def __init__(self, parent):
        """Init class."""
        from ..ui.python_kernel_spec_options import Ui_Form  # pylint: disable=import-outside-toplevel

        super().__init__(parent)
        self.ui = Ui_Form()
        self.ui.setupUi(self)
        self.kernel_spec_model = QStandardItemModel(self)
        self.ui.comboBox_kernel_specs.setModel(self.kernel_spec_model)
        self._refresh_kernel_spec_model()
        self._kernel_spec_editor = None
        # Initialize UI elements with defaults
        use_jupyter_console = int(self._toolbox.qsettings().value("appSettings/usePythonKernel", defaultValue="0"))
        if use_jupyter_console == 2:
            self.ui.radioButton_jupyter_console.setChecked(True)
        else:
            self.ui.radioButton_python_console.setChecked(True)
        default_python_path = self._toolbox.qsettings().value("appSettings/pythonPath", defaultValue="")
        self.ui.lineEdit_python_path.setPlaceholderText(resolve_python_interpreter(""))
        self.ui.lineEdit_python_path.setText(default_python_path)
        default_kernel_spec = self._toolbox.qsettings().value("appSettings/pythonKernel", defaultValue="")
        row = self.find_index_by_data(default_kernel_spec)
        if row == -1:
            self.ui.comboBox_kernel_specs.setCurrentIndex(0)
        else:
            self.ui.comboBox_kernel_specs.setCurrentIndex(row)
        self.set_ui_for_jupyter_console(not use_jupyter_console)
        self.connect_signals()

    def connect_signals(self):
        """Connects signals."""
        self.ui.toolButton_refresh_kernel_specs.clicked.connect(self._refresh_kernel_spec_model)
        self.ui.comboBox_kernel_specs.activated.connect(self._parent._push_change_kernel_spec_command)
        self.ui.radioButton_jupyter_console.toggled.connect(self._parent._push_set_jupyter_console_mode)
        self.ui.toolButton_browse_python.clicked.connect(self.browse_python_button_clicked)
        self.ui.pushButton_open_kernel_spec_viewer.clicked.connect(self.show_python_kernel_spec_editor)
        self.ui.lineEdit_python_path.editingFinished.connect(self._parent._push_change_executable)

    def init_widget(self, specification):
        """Initializes UI elements based on specification

        Args:
            specification (ToolSpecification): Specification to load
        """
        use_jupyter_console = specification.execution_settings["use_jupyter_console"]
        self.ui.radioButton_jupyter_console.blockSignals(True)
        self.ui.radioButton_python_console.blockSignals(True)
        if use_jupyter_console:
            self.ui.radioButton_jupyter_console.setChecked(True)
        else:
            self.ui.radioButton_python_console.setChecked(True)
        self.ui.radioButton_jupyter_console.blockSignals(False)
        self.ui.radioButton_python_console.blockSignals(False)
        self.set_ui_for_jupyter_console(not use_jupyter_console)
        k_spec = specification.execution_settings["kernel_spec_name"]
        if k_spec == "":
            self.ui.comboBox_kernel_specs.setCurrentIndex(0)  # Set 'Select kernel spec...'
        else:
            row = self.find_index_by_data(k_spec)
            if row == -1:
                notification = Notification(
                    self._parent, f"This Tool spec has kernel spec '{k_spec}' saved " f"but it could not be found."
                )
                notification.show()
                # TODO: What to do when a kernel spec name that is saved to Tool spec is not found?
                row += 1  # Set 'Select kernel spec...'
            self.ui.comboBox_kernel_specs.setCurrentIndex(row)
        self.set_executable(specification.execution_settings["executable"])

    def add_execution_settings(self):
        """Collects execution settings based on optional widget state into a dictionary, which is returned."""
        idx = self.ui.comboBox_kernel_specs.currentIndex()
        if idx < 1:
            d = dict()
            d["kernel_spec_name"] = ""
            d["env"] = ""
        else:
            item = self.ui.comboBox_kernel_specs.model().item(idx)
            k_spec_data = item.data()
            d = k_spec_data
        d["use_jupyter_console"] = self.ui.radioButton_jupyter_console.isChecked()
        self.validate_executable()  # Raises NameError if Python path is not valid
        d["executable"] = self.get_executable()
        return d

    @Slot(bool)
    def browse_python_button_clicked(self, _=False):
        """Calls static method that shows a file browser for selecting a Python interpreter."""
        select_python_interpreter(self, self.ui.lineEdit_python_path)
        self._parent._push_change_executable()

    def set_ui_for_jupyter_console(self, value):
        """Enables or disables some UI elements in the optional widget according to checkBox state."""
        self.ui.comboBox_kernel_specs.setDisabled(value)
        self.ui.toolButton_refresh_kernel_specs.setDisabled(value)
        self.ui.pushButton_open_kernel_spec_viewer.setDisabled(value)
        self.ui.lineEdit_python_path.setDisabled((not value))
        self.ui.toolButton_browse_python.setDisabled(not value)

    def validate_executable(self):
        """Check that Python path in the line edit is a file it exists and the file name starts with 'python'.

        Raises:
            NameError: If the python path in the line edit is not valid
        """
        p = self.ui.lineEdit_python_path.text().strip()
        if not file_is_valid(self._parent, p, "Invalid Python Interpreter", extra_check="python"):
            raise NameError

    def set_executable(self, p):
        self.ui.lineEdit_python_path.setText(p)

    def get_executable(self):
        return self.ui.lineEdit_python_path.text().strip()

    @Slot(bool)
    def show_python_kernel_spec_editor(self, _=False):
        """Opens kernel editor, where user can make kernel specs for the Jupyter Console."""
        p = self.ui.lineEdit_python_path.text()  # This may be an empty string
        j = ""
        item = self.kernel_spec_model.item(self.ui.comboBox_kernel_specs.currentIndex())
        selected_kernel_spec = item.data()["kernel_spec_name"]
        self._kernel_spec_editor = KernelEditor(self._parent, p, j, "python", selected_kernel_spec)
        self._kernel_spec_editor.finished.connect(self.python_kernel_editor_closed)
        self._kernel_spec_editor.open()

    @Slot(int)
    def python_kernel_editor_closed(self, ret_code):
        """Catches the selected Python kernel name when the editor is closed."""
        item = self.kernel_spec_model.item(self.ui.comboBox_kernel_specs.currentIndex())
        previous_kernel_spec = item.data()["kernel_spec_name"]
        self._refresh_kernel_spec_model()
        if ret_code != 1:  # Editor closed with Cancel
            # TODO: Is this needed?
            # Set previous kernel selected in Python kernel combobox if it still exists
            r = self.find_index_by_data(previous_kernel_spec)
            if r == -1:
                self.ui.comboBox_kernel_specs.setCurrentIndex(0)  # Previous not found
            else:
                self.ui.comboBox_kernel_specs.setCurrentIndex(r)
            return
        new_kernel_spec = self._kernel_spec_editor.selected_kernel
        row = self.find_index_by_data(new_kernel_spec)
        if row == -1:  # New kernel spec not found, should be quite uncommon
            notification = Notification(self, f"Python kernel spec {new_kernel_spec} not found")
            notification.show()
            self.ui.comboBox_kernel_specs.setCurrentIndex(0)
        else:
            self._parent._push_change_kernel_spec_command(row)

    def find_index_by_data(self, string):
        """Searches the kernel spec model for the first item whose data matches the given string.
        Returns the items row number or -1 if not found."""
        if string == "":
            return 0
        for row in range(1, self.kernel_spec_model.rowCount()):  # Start from index 1
            index = self.kernel_spec_model.index(row, 0)
            item = self.kernel_spec_model.itemFromIndex(index)
            d = item.data()
            if d["kernel_spec_name"] == string:
                return row
        return -1

    @busy_effect
    @Slot(bool)
    def _refresh_kernel_spec_model(self):
        item = self.kernel_spec_model.item(self.ui.comboBox_kernel_specs.currentIndex())
        if not item or not item.data():
            selected_kernel_spec = None
        else:
            selected_kernel_spec = item.data()["kernel_spec_name"]  # Remember the selected kernel spec
        self.kernel_spec_model.clear()
        first_item = QStandardItem("Select kernel spec...")
        self.kernel_spec_model.appendRow(first_item)
        # Add Python kernel specs
        kernel_specs = find_python_kernels()
        for n in kernel_specs.keys():
            item = QStandardItem(n + " [Jupyter]")
            spec_data = dict()
            spec_data["kernel_spec_name"] = n
            spec_data["env"] = ""
            item.setData(spec_data)
            self.kernel_spec_model.appendRow(item)
        # Add auto-generated conda kernel spec names
        conda_exe = self._toolbox.qsettings().value("appSettings/condaPath", defaultValue="")
        conda_exe = resolve_conda_executable(conda_exe)
        if conda_exe != "":
            ksm = CondaKernelSpecManager(conda_exe=conda_exe)
            conda_specs = ksm._all_specs()
            for i in conda_specs.keys():
                item = QStandardItem(i + " [Conda]")
                spec_data = dict()
                spec_data["kernel_spec_name"] = i
                spec_data["env"] = "conda"
                item.setData(spec_data)
                self.kernel_spec_model.appendRow(item)
        # Set the previously selected kernel spec as the current item after the model has been rebuilt
        if selected_kernel_spec is not None:
            row = self.find_index_by_data(selected_kernel_spec)
            if row == -1:
                # The kernel spec may have been removed
                self._parent._push_change_kernel_spec_command(0)
                return
            self.ui.comboBox_kernel_specs.setCurrentIndex(row)


class ExecutableToolSpecOptionalWidget(OptionalWidget):
    def __init__(self, parent):
        """Init class."""
        from ..ui.executable_cmd_exec_options import Ui_Form  # pylint: disable=import-outside-toplevel

        super().__init__(parent)
        self.ui = Ui_Form()
        self.ui.setupUi(self)
        self.shells = ["No shell", "cmd.exe", "powershell.exe", "bash"]
        self.ui.comboBox_shell.addItems(self.shells)
        self.connect_signals()

    def connect_signals(self):
        self.ui.lineEdit_command.editingFinished.connect(self._parent._push_change_executable_command)
        self.ui.comboBox_shell.activated.connect(self._parent._push_change_shell_command)

    def init_widget(self, specification):
        """Initializes UI elements based on specification."""
        self.ui.lineEdit_command.setText(specification.execution_settings["cmd"])
        shell = specification.execution_settings["shell"]
        ind = next(iter(k for k, t in enumerate(self.shells) if t.lower() == shell), 0)
        self.ui.comboBox_shell.setCurrentIndex(ind)

    def add_execution_settings(self):
        """Collects execution settings based on optional widget state into a dictionary, which is returned."""
        d = dict()
        d["cmd"] = self.ui.lineEdit_command.text()
        d["shell"] = self.get_current_shell()
        return d

    def get_current_shell(self):
        """Returns the selected shell in the shell combo box."""
        ind = self.ui.comboBox_shell.currentIndex()
        if ind < 1:
            return ""
        return self.ui.comboBox_shell.currentText()



# -*- coding: utf-8 -*-
######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

################################################################################
## Form generated from reading UI file 'url_selector_widget.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *

from spinetoolbox.widgets.custom_qlineedits import PropertyQLineEdit
from spinetoolbox.widgets.custom_qlineedits import CustomQLineEdit

from spine_items import resources_icons_rc

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(363, 279)
        self.verticalLayout = QVBoxLayout(Dialog)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2 = QVBoxLayout()
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(9, 9, 9, 9)
        self.frame = QFrame(Dialog)
        self.frame.setObjectName(u"frame")
        self.gridLayout = QGridLayout(self.frame)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(3, 3, 3, 3)
        self.lineEdit_password = PropertyQLineEdit(self.frame)
        self.lineEdit_password.setObjectName(u"lineEdit_password")
        self.lineEdit_password.setEnabled(False)
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lineEdit_password.sizePolicy().hasHeightForWidth())
        self.lineEdit_password.setSizePolicy(sizePolicy)
        self.lineEdit_password.setMinimumSize(QSize(0, 24))
        self.lineEdit_password.setMaximumSize(QSize(5000, 24))
        self.lineEdit_password.setEchoMode(QLineEdit.Password)
        self.lineEdit_password.setClearButtonEnabled(True)

        self.gridLayout.addWidget(self.lineEdit_password, 3, 1, 1, 5)

        self.label_port = QLabel(self.frame)
        self.label_port.setObjectName(u"label_port")
        sizePolicy1 = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy1.setHorizontalStretch(1)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.label_port.sizePolicy().hasHeightForWidth())
        self.label_port.setSizePolicy(sizePolicy1)

        self.gridLayout.addWidget(self.label_port, 4, 4, 1, 1)

        self.comboBox_dialect = QComboBox(self.frame)
        self.comboBox_dialect.setObjectName(u"comboBox_dialect")
        sizePolicy.setHeightForWidth(self.comboBox_dialect.sizePolicy().hasHeightForWidth())
        self.comboBox_dialect.setSizePolicy(sizePolicy)
        self.comboBox_dialect.setMinimumSize(QSize(0, 24))
        self.comboBox_dialect.setMaximumSize(QSize(16777215, 24))

        self.gridLayout.addWidget(self.comboBox_dialect, 0, 1, 1, 5)

        self.label_dsn = QLabel(self.frame)
        self.label_dsn.setObjectName(u"label_dsn")
        sizePolicy2 = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.label_dsn.sizePolicy().hasHeightForWidth())
        self.label_dsn.setSizePolicy(sizePolicy2)

        self.gridLayout.addWidget(self.label_dsn, 1, 0, 1, 1)

        self.label_dialect = QLabel(self.frame)
        self.label_dialect.setObjectName(u"label_dialect")
        sizePolicy2.setHeightForWidth(self.label_dialect.sizePolicy().hasHeightForWidth())
        self.label_dialect.setSizePolicy(sizePolicy2)
        self.label_dialect.setMaximumSize(QSize(16777215, 16777215))

        self.gridLayout.addWidget(self.label_dialect, 0, 0, 1, 1)

        self.lineEdit_port = PropertyQLineEdit(self.frame)
        self.lineEdit_port.setObjectName(u"lineEdit_port")
        self.lineEdit_port.setEnabled(False)
        sizePolicy.setHeightForWidth(self.lineEdit_port.sizePolicy().hasHeightForWidth())
        self.lineEdit_port.setSizePolicy(sizePolicy)
        self.lineEdit_port.setMinimumSize(QSize(0, 24))
        self.lineEdit_port.setMaximumSize(QSize(80, 24))
        self.lineEdit_port.setInputMethodHints(Qt.ImhNone)

        self.gridLayout.addWidget(self.lineEdit_port, 4, 5, 1, 1)

        self.label_username = QLabel(self.frame)
        self.label_username.setObjectName(u"label_username")
        sizePolicy2.setHeightForWidth(self.label_username.sizePolicy().hasHeightForWidth())
        self.label_username.setSizePolicy(sizePolicy2)

        self.gridLayout.addWidget(self.label_username, 2, 0, 1, 1)

        self.label_database = QLabel(self.frame)
        self.label_database.setObjectName(u"label_database")
        sizePolicy2.setHeightForWidth(self.label_database.sizePolicy().hasHeightForWidth())
        self.label_database.setSizePolicy(sizePolicy2)

        self.gridLayout.addWidget(self.label_database, 6, 0, 1, 1)

        self.comboBox_dsn = QComboBox(self.frame)
        self.comboBox_dsn.setObjectName(u"comboBox_dsn")
        self.comboBox_dsn.setEnabled(False)
        sizePolicy.setHeightForWidth(self.comboBox_dsn.sizePolicy().hasHeightForWidth())
        self.comboBox_dsn.setSizePolicy(sizePolicy)
        self.comboBox_dsn.setMinimumSize(QSize(0, 24))
        self.comboBox_dsn.setMaximumSize(QSize(16777215, 24))

        self.gridLayout.addWidget(self.comboBox_dsn, 1, 1, 1, 5)

        self.lineEdit_host = PropertyQLineEdit(self.frame)
        self.lineEdit_host.setObjectName(u"lineEdit_host")
        self.lineEdit_host.setEnabled(False)
        sizePolicy3 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy3.setHorizontalStretch(3)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.lineEdit_host.sizePolicy().hasHeightForWidth())
        self.lineEdit_host.setSizePolicy(sizePolicy3)
        self.lineEdit_host.setMinimumSize(QSize(0, 24))
        self.lineEdit_host.setMaximumSize(QSize(5000, 24))
        self.lineEdit_host.setClearButtonEnabled(True)

        self.gridLayout.addWidget(self.lineEdit_host, 4, 1, 1, 3)

        self.label_password = QLabel(self.frame)
        self.label_password.setObjectName(u"label_password")
        sizePolicy2.setHeightForWidth(self.label_password.sizePolicy().hasHeightForWidth())
        self.label_password.setSizePolicy(sizePolicy2)

        self.gridLayout.addWidget(self.label_password, 3, 0, 1, 1)

        self.lineEdit_username = PropertyQLineEdit(self.frame)
        self.lineEdit_username.setObjectName(u"lineEdit_username")
        self.lineEdit_username.setEnabled(False)
        sizePolicy.setHeightForWidth(self.lineEdit_username.sizePolicy().hasHeightForWidth())
        self.lineEdit_username.setSizePolicy(sizePolicy)
        self.lineEdit_username.setMinimumSize(QSize(0, 24))
        self.lineEdit_username.setMaximumSize(QSize(5000, 24))
        self.lineEdit_username.setClearButtonEnabled(True)

        self.gridLayout.addWidget(self.lineEdit_username, 2, 1, 1, 5)

        self.label_host = QLabel(self.frame)
        self.label_host.setObjectName(u"label_host")
        sizePolicy2.setHeightForWidth(self.label_host.sizePolicy().hasHeightForWidth())
        self.label_host.setSizePolicy(sizePolicy2)

        self.gridLayout.addWidget(self.label_host, 4, 0, 1, 1)

        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.lineEdit_database = CustomQLineEdit(self.frame)
        self.lineEdit_database.setObjectName(u"lineEdit_database")
        self.lineEdit_database.setEnabled(False)
        sizePolicy.setHeightForWidth(self.lineEdit_database.sizePolicy().hasHeightForWidth())
        self.lineEdit_database.setSizePolicy(sizePolicy)
        self.lineEdit_database.setMinimumSize(QSize(0, 24))
        self.lineEdit_database.setMaximumSize(QSize(16777215, 24))
        self.lineEdit_database.setCursor(QCursor(Qt.IBeamCursor))
        self.lineEdit_database.setClearButtonEnabled(True)

        self.horizontalLayout_4.addWidget(self.lineEdit_database)

        self.toolButton_select_sqlite_file = QToolButton(self.frame)
        self.toolButton_select_sqlite_file.setObjectName(u"toolButton_select_sqlite_file")
        self.toolButton_select_sqlite_file.setEnabled(False)
        sizePolicy.setHeightForWidth(self.toolButton_select_sqlite_file.sizePolicy().hasHeightForWidth())
        self.toolButton_select_sqlite_file.setSizePolicy(sizePolicy)
        self.toolButton_select_sqlite_file.setMinimumSize(QSize(22, 22))
        self.toolButton_select_sqlite_file.setMaximumSize(QSize(22, 22))
        icon = QIcon()
        icon.addFile(u":/icons/folder-open-solid.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.toolButton_select_sqlite_file.setIcon(icon)

        self.horizontalLayout_4.addWidget(self.toolButton_select_sqlite_file)


        self.gridLayout.addLayout(self.horizontalLayout_4, 6, 1, 1, 5)


        self.verticalLayout_2.addWidget(self.frame)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.verticalLayout_2.addItem(self.verticalSpacer)

        self.buttonBox = QDialogButtonBox(Dialog)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.verticalLayout_2.addWidget(self.buttonBox)


        self.verticalLayout.addLayout(self.verticalLayout_2)

        self.horizontalLayout_statusbar_ph = QHBoxLayout()
        self.horizontalLayout_statusbar_ph.setObjectName(u"horizontalLayout_statusbar_ph")
        self.widget_invisible_dummy = QWidget(Dialog)
        self.widget_invisible_dummy.setObjectName(u"widget_invisible_dummy")

        self.horizontalLayout_statusbar_ph.addWidget(self.widget_invisible_dummy)


        self.verticalLayout.addLayout(self.horizontalLayout_statusbar_ph)

        QWidget.setTabOrder(self.comboBox_dialect, self.comboBox_dsn)
        QWidget.setTabOrder(self.comboBox_dsn, self.lineEdit_username)
        QWidget.setTabOrder(self.lineEdit_username, self.lineEdit_password)
        QWidget.setTabOrder(self.lineEdit_password, self.lineEdit_host)
        QWidget.setTabOrder(self.lineEdit_host, self.lineEdit_port)
        QWidget.setTabOrder(self.lineEdit_port, self.lineEdit_database)
        QWidget.setTabOrder(self.lineEdit_database, self.toolButton_select_sqlite_file)

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)

        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"Select URL", None))
        self.lineEdit_password.setPlaceholderText("")
        self.label_port.setText(QCoreApplication.translate("Dialog", u"Port", None))
        self.label_dsn.setText(QCoreApplication.translate("Dialog", u"DSN", None))
        self.label_dialect.setText(QCoreApplication.translate("Dialog", u"Dialect", None))
        self.lineEdit_port.setPlaceholderText("")
        self.label_username.setText(QCoreApplication.translate("Dialog", u"Username", None))
        self.label_database.setText(QCoreApplication.translate("Dialog", u"Database", None))
        self.lineEdit_host.setPlaceholderText("")
        self.label_password.setText(QCoreApplication.translate("Dialog", u"Password", None))
        self.lineEdit_username.setPlaceholderText("")
        self.label_host.setText(QCoreApplication.translate("Dialog", u"Host", None))
        self.lineEdit_database.setPlaceholderText("")
#if QT_CONFIG(tooltip)
        self.toolButton_select_sqlite_file.setToolTip(QCoreApplication.translate("Dialog", u"<html><head/><body><p>Select SQLite file</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
    # retranslateUi




######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Module for data connection class.

:author: P. Savolainen (VTT)
:date:   19.12.2017
"""

import os
import shutil
import logging
from sqlalchemy import create_engine
from PySide2.QtCore import Slot, Qt, QFileInfo, QModelIndex, QItemSelection
from PySide2.QtGui import QStandardItem, QStandardItemModel, QBrush
from PySide2.QtWidgets import QFileDialog, QGraphicsItem, QFileIconProvider, QInputDialog, QMessageBox
from spine_engine.utils.serialization import deserialize_path, serialize_path
from spinetoolbox.project_item.project_item import ProjectItem
from spinetoolbox.helpers import open_url
from spinetoolbox.config import INVALID_FILENAME_CHARS
from .commands import AddDCReferencesCommand, RemoveDCReferencesCommand, MoveReferenceToData
from .custom_file_system_watcher import CustomFileSystemWatcher
from .executable_item import ExecutableItem
from .item_info import ItemInfo
from .output_resources import scan_for_resources
from ..widgets import UrlSelector
from ..utils import split_url_credentials


class DataConnection(ProjectItem):
    def __init__(
        self, name, description, x, y, toolbox, project, file_references=None, db_references=None, db_credentials=None
    ):
        """Data Connection class.

        Args:
            name (str): Object name
            description (str): Object description
            x (float): Initial X coordinate of item icon
            y (float): Initial Y coordinate of item icon
            toolbox (ToolboxUI): QMainWindow instance
            project (SpineToolboxProject): the project this item belongs to
            file_references (list, optional): a list of file paths
            db_references (list, optional): a list of db urls
            db_credentials (dict, optional): mapping urls from db_references to tuple (username, password)
        """
        super().__init__(name, description, x, y, project)
        if file_references is None:
            file_references = list()
        if db_references is None:
            db_references = list()
        if db_credentials is None:
            db_credentials = dict()
        self._toolbox = toolbox
        self.reference_model = QStandardItemModel()  # References to files
        self.data_model = QStandardItemModel()  # Paths of project internal files. These are found in DC data directory
        self.file_system_watcher = None
        self.file_references = list(file_references)
        self.db_references = list(db_references)
        self.db_credentials = db_credentials
        self._file_ref_root = None
        self._db_ref_root = None
        self.file_refs_selected = False
        self.any_refs_selected = False
        self.any_data_selected = False
        self.populate_reference_list()
        self.populate_data_list()

    def set_up(self):
        super().set_up()
        self.file_system_watcher = CustomFileSystemWatcher(self)
        self.file_system_watcher.add_persistent_file_paths(ref for ref in self.file_references if os.path.exists(ref))
        self.file_system_watcher.add_persistent_dir_path(self.data_dir)
        self.file_system_watcher.file_removed.connect(self._handle_file_removed)
        self.file_system_watcher.file_renamed.connect(self._handle_file_renamed)
        self.file_system_watcher.file_added.connect(self._handle_file_added)

    @staticmethod
    def item_type():
        """See base class."""
        return ItemInfo.item_type()

    @staticmethod
    def item_category():
        """See base class."""
        return ItemInfo.item_category()

    @property
    def executable_class(self):
        return ExecutableItem

    @Slot(QItemSelection, QItemSelection)
    def _update_selection_state(self, _selected, _deselected):
        self._do_update_selection_state()

    def _do_update_selection_state(self):
        ref_indexes = self._properties_ui.treeView_dc_references.selectionModel().selectedIndexes()
        data_indexes = self._properties_ui.treeView_dc_data.selectionModel().selectedIndexes()
        self.file_refs_selected = any(ind.parent().row() == 0 for ind in ref_indexes)
        self.any_refs_selected = any(ind.parent().row() in (0, 1) for ind in ref_indexes)
        self.current_is_file_ref = self._properties_ui.treeView_dc_references.currentIndex().parent().row() == 0
        self.any_data_selected = bool(data_indexes)
        self._properties_ui.toolButton_minus.setEnabled(self.any_refs_selected)
        self._properties_ui.toolButton_add.setEnabled(self.file_refs_selected)

    def make_signal_handler_dict(self):
        """Returns a dictionary of all shared signals and their handlers.
        This is to enable simpler connecting and disconnecting."""
        s = super().make_signal_handler_dict()
        # pylint: disable=unnecessary-lambda
        s[self._properties_ui.toolButton_plus_file.clicked] = self.show_add_file_references_dialog
        s[self._properties_ui.toolButton_plus_url.clicked] = self.show_add_db_reference_dialog
        s[self._properties_ui.toolButton_minus.clicked] = self.remove_references
        s[self._properties_ui.toolButton_add.clicked] = self.copy_to_project
        s[self._properties_ui.treeView_dc_references.doubleClicked] = self.open_reference
        s[self._properties_ui.treeView_dc_data.doubleClicked] = self.open_data_file
        s[self._properties_ui.treeView_dc_references.files_dropped] = self._add_file_references
        s[self._properties_ui.treeView_dc_data.files_dropped] = self.add_data_files
        s[self.get_icon().files_dropped_on_icon] = self.receive_files_dropped_on_icon
        s[self._properties_ui.treeView_dc_references.del_key_pressed] = lambda: self.remove_references
        s[self._properties_ui.treeView_dc_data.del_key_pressed] = self.remove_files
        s[self._properties_ui.treeView_dc_references.selectionModel().selectionChanged] = self._update_selection_state
        s[self._properties_ui.treeView_dc_data.selectionModel().selectionChanged] = self._update_selection_state
        return s

    def restore_selections(self):
        """Restore selections into shared widgets when this project item is selected."""
        self._properties_ui.treeView_dc_references.setModel(self.reference_model)
        self._properties_ui.treeView_dc_data.setModel(self.data_model)
        self._do_update_selection_state()
        self._properties_ui.treeView_dc_references.expandAll()

    @Slot(QGraphicsItem, list)
    def receive_files_dropped_on_icon(self, icon, file_paths):
        """Called when files are dropped onto a data connection graphics item.
        If the item is this Data Connection's graphics item, add the files to data."""
        if icon == self.get_icon():
            self.add_data_files(file_paths)

    @Slot(list)
    def add_data_files(self, file_paths):
        """Add files to data directory"""
        for file_path in file_paths:
            filename = os.path.split(file_path)[1]
            self._logger.msg.emit(f"Copying file <b>{filename}</b> to <b>{self.name}</b>")
            try:
                shutil.copy(file_path, self.data_dir)
            except OSError:
                self._logger.msg_error.emit("[OSError] Copying failed")

    @Slot(bool)
    def show_add_file_references_dialog(self, checked=False):
        """Opens a file browser where user can select files to be added as references for this Data Connection."""
        answer = QFileDialog.getOpenFileNames(self._toolbox, "Add file references", self._project.project_dir, "*.*")
        file_paths = answer[0]
        if not file_paths:  # Cancel button clicked
            return
        self._add_file_references(file_paths)

    @Slot(list)
    def _add_file_references(self, paths):
        """Add multiple file paths to reference list.

        Args:
            paths (list): A list of paths to files
        """
        repeated_paths = []
        new_paths = []
        for path in paths:
            if not os.path.isfile(path):
                continue
            if any(os.path.samefile(path, ref) for ref in self.file_references):
                repeated_paths.append(path)
            else:
                new_paths.append(path)
        repeated_paths = ", ".join(repeated_paths)
        if repeated_paths:
            self._logger.msg_warning.emit(f"Reference to file(s) <b>{repeated_paths}</b> already exists")
        if new_paths:
            self._toolbox.undo_stack.push(AddDCReferencesCommand(self, new_paths, []))

    @Slot(bool)
    def show_add_db_reference_dialog(self, checked=False):
        """Opens a dialog where user can select a url to be added as reference for this Data Connection."""
        selector = UrlSelector(self._toolbox, self._toolbox)
        selector.exec_()
        url = selector.url
        if not url:  # Cancel button clicked
            return
        try:
            engine = create_engine(url)
            with engine.connect():
                pass
        except Exception as e:  # pylint: disable=broad-except
            self._logger.msg_error.emit(f"Unable to connect to <b>{url}</b>: {e}")
            return
        url, credentials = split_url_credentials(url)
        if url in self.db_references:
            self._logger.msg_warning.emit(f"Reference to database <b>{url}</b> already exists")
            return
        self.db_credentials[url] = credentials
        self._toolbox.undo_stack.push(AddDCReferencesCommand(self, [], [url]))

    def do_add_references(self, file_refs, db_refs):
        file_refs = [os.path.abspath(ref) for ref in file_refs]
        self.file_references += file_refs
        self.file_system_watcher.add_persistent_file_paths(ref for ref in file_refs if os.path.exists(ref))
        self._append_file_references_to_model(*file_refs)
        self.db_references += db_refs
        self._append_db_references_to_model(*db_refs)
        self._check_notifications()
        self._resources_to_successors_changed()

    @Slot(bool)
    def remove_references(self, checked=False):
        """Pushes a remove references command to undo stack"""
        indexes = self._properties_ui.treeView_dc_references.selectedIndexes()
        if not indexes:  # Nothing selected
            self._logger.msg.emit("Please select references to remove")
            return
        file_ref_root_index = self.reference_model.indexFromItem(self._file_ref_root)
        db_ref_root_index = self.reference_model.indexFromItem(self._db_ref_root)
        file_references = []
        db_references = []
        for index in indexes:
            parent = index.parent()
            if not parent.isValid():
                continue
            if parent == file_ref_root_index:
                file_references.append(index.data(Qt.DisplayRole))
            elif parent == db_ref_root_index:
                db_references.append(index.data(Qt.DisplayRole))
        self._toolbox.undo_stack.push(RemoveDCReferencesCommand(self, file_references, db_references))
        self._logger.msg.emit("Selected references removed")

    def do_remove_references(self, file_refs, db_refs):
        """Removes given paths from references.

        Args:
            file_refs (list): List of removed file paths.
            db_refs (list): List of removed urls.
        """
        self.file_system_watcher.remove_persistent_file_paths(file_refs)
        refs_removed = self._remove_file_references(*file_refs)
        refs_removed |= self._remove_db_references(*db_refs)
        if refs_removed:
            self._check_notifications()
            self._resources_to_successors_changed()

    def _remove_file_references(self, *refs):
        result = False
        for k in reversed(range(self._file_ref_root.rowCount())):
            if any(_samepath(self._file_ref_root.child(k).text(), ref) for ref in refs):
                self.file_references.pop(k)
                self._file_ref_root.removeRow(k)
                result = True
        return result

    def _remove_db_references(self, *refs):
        result = False
        for k in reversed(range(self._db_ref_root.rowCount())):
            if self._db_ref_root.child(k).text() in refs:
                url = self.db_references.pop(k)
                self.db_credentials.pop(url, None)
                self._db_ref_root.removeRow(k)
                result = True
        return result

    def _rename_file_reference(self, old_path, new_path):
        for k in range(self._file_ref_root.rowCount()):
            item = self._file_ref_root.child(k)
            if _samepath(item.text(), old_path):
                item.setText(new_path)
                self.file_references[k] = new_path
                return True
        return False

    def _remove_data_file(self, path):
        for k in reversed(range(self.data_model.rowCount())):
            data_filepath = self.data_model.item(k).data(Qt.UserRole)
            if _samepath(data_filepath, path):
                self.data_model.removeRow(k)
                return True
        return False

    def _rename_data_file(self, old_path, new_path):
        for k in range(self.data_model.rowCount()):
            item = self.data_model.item(k)
            if _samepath(item.data(Qt.UserRole), old_path):
                item.setText(os.path.basename(new_path))
                item.setData(new_path, Qt.UserRole)
                return True
        return False

    @Slot(str)
    def _handle_file_removed(self, path):
        if self._remove_file_references(path) or self._remove_data_file(path):
            self._check_notifications()
            self._resources_to_successors_changed()

    @Slot(str, str)
    def _handle_file_renamed(self, old_path, new_path):
        def replace_new_path(paths):
            for i, path in enumerate(paths):
                if path == new_path:
                    paths[i] = old_path
                    return True
            return False

        renamed = self._rename_file_reference(old_path, new_path)
        if not renamed:
            renamed = self._rename_data_file(old_path, new_path)
        if not renamed:
            return
        self._check_notifications()
        file_refs = list(self.file_references)
        data_files = [os.path.join(self.data_dir, f) for f in self.data_files()]
        new_resources = scan_for_resources(
            self, file_refs + data_files, self.db_references, self.db_credentials, self._project.project_dir
        )
        if not replace_new_path(file_refs):
            replace_new_path(data_files)
        old_resources = scan_for_resources(
            self, file_refs + data_files, self.db_references, self.db_credentials, self._project.project_dir
        )
        self._resources_to_successors_replaced(old_resources, new_resources)

    @Slot(str)
    def _handle_file_added(self, path):
        if _samepath(os.path.dirname(path), self.data_dir):
            self._append_data_files_to_model(path)
            self._check_notifications()
            self._resources_to_successors_changed()

    @Slot(bool)
    def copy_to_project(self, checked=False):
        """Copy selected file references to this Data Connection's data directory."""
        selected_indexes = self._properties_ui.treeView_dc_references.selectedIndexes()
        if not selected_indexes:
            self._logger.msg_warning.emit("No files to copy")
            return
        self._toolbox.undo_stack.push(MoveReferenceToData(self, [index.data() for index in selected_indexes]))

    def do_copy_to_project(self, paths):
        """Copies given files to item's data directory.

        Args:
            paths (Iterable of str): paths to copy
        """
        for path in paths:
            if not os.path.exists(path):
                self._logger.msg_error.emit(f"File <b>{path}</b> does not exist")
                continue
            filename = os.path.basename(path)
            self._logger.msg.emit(f"Copying file <b>{filename}</b> to Data Connection <b>{self.name}</b>")
            try:
                shutil.copy(path, self.data_dir)
            except OSError:
                self._logger.msg_error.emit("[OSError] Copying failed")

    @Slot(QModelIndex)
    def open_reference(self, index):
        """Open reference in default program."""
        if not index:
            return
        if not index.isValid():
            logging.error("Index not valid")
            return
        parent_item = self.reference_model.itemFromIndex(index.parent())
        if parent_item is not self._file_ref_root:
            return
        reference = self.file_references[index.row()]
        url = "file:///" + reference
        res = open_url(url)
        if not res:
            self._logger.msg_error.emit(f"Failed to open reference:<b>{reference}</b>")

    @Slot(QModelIndex)
    def open_data_file(self, index):
        """Open data file in default program."""
        if not index:
            return
        if not index.isValid():
            logging.error("Index not valid")
            return
        data_file = index.data(Qt.UserRole)
        url = "file:///" + data_file
        # noinspection PyTypeChecker, PyCallByClass, PyArgumentList
        res = open_url(url)
        if not res:
            self._logger.msg_error.emit(f"Opening file <b>{data_file}</b> failed")

    def make_new_file(self):
        """Create a new blank file to this Data Connections data directory."""
        msg = "File name"
        # noinspection PyCallByClass, PyTypeChecker, PyArgumentList
        answer = QInputDialog.getText(
            self._toolbox, "Create new file", msg, flags=Qt.WindowTitleHint | Qt.WindowCloseButtonHint
        )
        file_name = answer[0]
        if not file_name.strip():
            return
        # Check that file name has no invalid chars
        if any(True for x in file_name if x in INVALID_FILENAME_CHARS):
            msg = f"File name <b>{file_name}</b> contains invalid characters."
            self._logger.information_box.emit("Creating file failed", msg)
            return
        file_path = os.path.join(self.data_dir, file_name)
        if os.path.exists(file_path):
            msg = f"File <b>{file_name}</b> already exists."
            self._logger.information_box.emit("Creating file failed", msg)
            return
        try:
            with open(file_path, "w"):
                self._logger.msg.emit(f"File <b>{file_name}</b> created to Data Connection <b>{self.name}</b>")
        except OSError:
            msg = "Please check directory permissions."
            self._logger.information_box.emit("Creating file failed", msg)

    @Slot()
    def remove_files(self):
        """Remove selected files from data directory."""
        indexes = self._properties_ui.treeView_dc_data.selectedIndexes()
        if not indexes:  # Nothing selected
            self._logger.msg.emit("Please select files to remove")
            return
        file_list = [index.data() for index in indexes]
        files = "\n".join(file_list)
        msg = "The following files will be removed permanently from the project\n\n" f"{files}\n\n" "Are you sure?"
        title = f"Remove {len(file_list)} File(s)"
        message_box = QMessageBox(
            QMessageBox.Question, title, msg, QMessageBox.Ok | QMessageBox.Cancel, parent=self._toolbox
        )
        message_box.button(QMessageBox.Ok).setText("Remove Files")
        answer = message_box.exec_()
        if answer == QMessageBox.Cancel:
            return
        self.delete_files_from_project(file_list)

    def delete_files_from_project(self, file_names):
        """Deletes given files from item's data directory.

        Args:
            file_names (Iterable of str): files to delete
        """
        for filename in file_names:
            path_to_remove = os.path.join(self.data_dir, filename)
            try:
                os.remove(path_to_remove)
                self._logger.msg.emit(f"File <b>{path_to_remove}</b> removed")
            except OSError:
                self._logger.msg_error.emit(f"Removing file {path_to_remove} failed.\nCheck permissions.")

    def populate_reference_list(self):
        """List file references in QTreeView."""
        self.reference_model.clear()
        self.reference_model.setHorizontalHeaderItem(0, QStandardItem("References"))  # Add header
        self._file_ref_root = QStandardItem("File paths")
        self._db_ref_root = QStandardItem("URLs")
        self._file_ref_root.setFlags(self._file_ref_root.flags() & ~Qt.ItemIsEditable)
        self._db_ref_root.setFlags(self._file_ref_root.flags() & ~Qt.ItemIsEditable)
        self.reference_model.appendRow(self._file_ref_root)
        self.reference_model.appendRow(self._db_ref_root)
        self._append_file_references_to_model(*self.file_references)
        self._append_db_references_to_model(*self.db_references)

    def _append_file_references_to_model(self, *paths):
        for path in paths:
            item = QStandardItem(path)
            item.setFlags(~Qt.ItemIsEditable)
            if not os.path.exists(path):
                self._logger.msg_error.emit(f"<b>{self.name}:</b> Could not find file reference {path}.")
                tooltip = "The file is missing."
                item.setData(tooltip, Qt.ToolTipRole)
                item.setData(QBrush(Qt.red), Qt.ForegroundRole)
            self._file_ref_root.appendRow(item)

    def _append_db_references_to_model(self, *urls):
        for url in urls:
            item = QStandardItem(url)
            item.setFlags(~Qt.ItemIsEditable)
            self._db_ref_root.appendRow(item)

    def populate_data_list(self):
        """List project internal data (files) in QTreeView."""
        self.data_model.clear()
        self.data_model.setHorizontalHeaderItem(0, QStandardItem("Data"))  # Add header
        self._append_data_files_to_model(*self.data_files())

    def _append_data_files_to_model(self, *paths):
        for path in paths:
            item = QStandardItem(os.path.basename(path))
            item.setFlags(~Qt.ItemIsEditable)
            icon = QFileIconProvider().icon(QFileInfo(path))
            item.setData(icon, Qt.DecorationRole)
            item.setData(path, Qt.UserRole)
            self.data_model.appendRow(item)

    def resources_for_direct_successors(self):
        """see base class"""
        data_files = [os.path.join(self.data_dir, f) for f in self.data_files()]
        resources = scan_for_resources(
            self, self.file_references + data_files, self.db_references, self.db_credentials, self._project.project_dir
        )
        return resources

    def _check_notifications(self):
        """Sets or clears the exclamation mark icon."""
        self.clear_notifications()
        if not self.file_references and not self.db_references and not self.data_files():
            self.add_notification(
                "This Data Connection does not have any references or data. "
                "Add some in the Data Connection Properties panel."
            )
        missing_file_references = [ref for ref in self.file_references if not os.path.exists(ref)]
        if missing_file_references:
            self.add_notification("Cannot find some file references. Please, check that the files exist.")

    def item_dict(self):
        """Returns a dictionary corresponding to this item."""
        d = super().item_dict()
        d["file_references"] = [serialize_path(ref, self._project.project_dir) for ref in self.file_references]
        d["db_references"] = self.db_references
        d["db_credentials"] = self.db_credentials
        return d

    @staticmethod
    def item_dict_local_entries():
        """See base class."""
        return [("db_credentials",)]

    @staticmethod
    def from_dict(name, item_dict, toolbox, project):
        description, x, y = ProjectItem.parse_item_dict(item_dict)
        # FIXME: Do we want to convert references to file_references via upgrade?
        file_references = item_dict.get("file_references", list()) or item_dict.get("references", list())
        file_references = [deserialize_path(r, project.project_dir) for r in file_references]
        db_references = item_dict.get("db_references", list())
        db_credentials = item_dict.get("db_credentials", dict())
        return DataConnection(name, description, x, y, toolbox, project, file_references, db_references, db_credentials)

    def rename(self, new_name, rename_data_dir_message):
        """See base class."""
        old_data_dir = self.data_dir
        if not super().rename(new_name, rename_data_dir_message):
            return False
        self.file_system_watcher.remove_persistent_dir_path(old_data_dir)
        self.file_system_watcher.add_persistent_dir_path(self.data_dir)
        self.populate_data_list()
        return True

    def tear_down(self):
        """Tears down this item. Called by toolbox just before closing."""
        super().tear_down()
        self.file_system_watcher.tear_down()

    def notify_destination(self, source_item):
        """See base class."""
        if source_item.item_type() == "Tool":
            self._logger.msg.emit(
                f"Link established. Tool <b>{source_item.name}</b> output files will be "
                f"passed as references to item <b>{self.name}</b> after execution."
            )
        elif source_item.item_type() in ["Data Store", "Importer"]:
            # Does this type of link do anything?
            self._logger.msg.emit("Link established")
        else:
            super().notify_destination(source_item)


def _samepath(path1, path2):
    return os.path.normcase(path1) == os.path.normcase(path2)



######################################################################################################################
# Copyright (C) 2017-2021 Spine project consortium
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""
Parameter merging settings widget.

:author: A. Soininen (VTT)
:date:   19.2.2020
"""

from PySide2.QtCore import QItemSelection, QItemSelectionModel, Signal, Slot
from PySide2.QtWidgets import QWidget
from spinedb_api.spine_io.exporters.gdx import MergingSetting
from .merging_error_flag import MergingErrorFlag
from ..mvcmodels.domain_name_list_model import DomainNameListModel
from ..mvcmodels.parameter_name_list_model import ParameterNameListModel

_ERROR_MESSAGE = "<span style='color:#ff3333;white-space: pre-wrap;'>{}</span>"


class ParameterMergingSettings(QWidget):
    """A widget for configuring parameter merging."""

    removal_requested = Signal("QVariant")
    """Emitted when the settings widget wants to get removed from the parent window."""

    def __init__(self, entity_class_infos, set_settings, parent, parameter_name=None, merging_setting=None):
        """
        Args:
            entity_class_infos (list): list of EntityClassInfo objects
            set_settings (SetSettings): set settings
            parent (QWidget): a parent widget
            parameter_name (str): merged parameter name of None for widget
            merging_setting (MergingSetting): merging settings or None for empty widget
        """
        from ..ui.parameter_merging_settings import Ui_Form  # pylint: disable=import-outside-toplevel

        super().__init__(parent)
        self._set_settings = set_settings
        self._error_flags = (
            MergingErrorFlag.DOMAIN_NAME_MISSING
            | MergingErrorFlag.PARAMETER_NAME_MISSING
            | MergingErrorFlag.NO_PARAMETER_SELECTED
        )
        self._parameter_name = ""
        self._selected_domain_row = None
        self._index_position = 0
        self._ui = Ui_Form()
        self._ui.setupUi(self)
        self._ui.parameter_name_edit.textChanged.connect(self._update_parameter_name)
        self._ui.remove_button.clicked.connect(self._remove_self)
        self._domain_names_model = DomainNameListModel(entity_class_infos)
        self._ui.domains_list_view.setModel(self._domain_names_model)
        self._ui.domains_list_view.selectionModel().selectionChanged.connect(self._handle_domain_selection_change)
        self._parameter_name_list_model = ParameterNameListModel([])
        self._ui.parameter_name_list_view.setModel(self._parameter_name_list_model)
        self._ui.domain_name_edit.textChanged.connect(self._update_indexing_domain_name)
        self._ui.move_domain_left_button.clicked.connect(self._move_domain_left)
        self._ui.move_domain_right_button.clicked.connect(self._move_domain_right)
        self._ui.message_label.setText("")
        if parameter_name is not None:
            self._ui.parameter_name_edit.setText(parameter_name)
        if merging_setting is not None:
            self._parameter_name_list_model.known_selections(
                merging_setting.previous_set, merging_setting.parameter_names
            )
            domain_index = self._domain_names_model.index_for(merging_setting.previous_set)
            self._ui.domains_list_view.selectionModel().select(domain_index, QItemSelectionModel.Select)
            self._ui.domain_name_edit.setText(merging_setting.new_domain_name)
            self._ui.domain_description_edit.setText(merging_setting.new_domain_description)
            self._index_position = merging_setting.index_position
            self._reset_indexing_domains_label()
        self._check_state()

    @property
    def error_flags(self):
        return self._error_flags

    @property
    def parameter_name(self):
        """Name of the merged parameter."""
        return self._parameter_name

    def merging_setting(self):
        """Constructs the MergingSetting object from the widget's contents."""
        parameter_names = self._parameter_name_list_model.selected()
        domain_name = self._ui.domain_name_edit.text()
        domain_description = self._ui.domain_description_edit.text()
        entity_class_info = self._domain_names_model.item_at(self._selected_domain_row)
        previous_set = entity_class_info.name
        previous_domain_names = entity_class_info.domain_names
        setting = MergingSetting(parameter_names, domain_name, domain_description, previous_set, previous_domain_names)
        setting.index_position = self._index_position
        return setting

    def _check_state(self):
        """Updates the message label according to widget's error state."""
        if self._error_flags & MergingErrorFlag.PARAMETER_NAME_MISSING:
            self._ui.message_label.setText(_ERROR_MESSAGE.format("Parameter name missing."))
        elif self._error_flags & MergingErrorFlag.DOMAIN_NAME_MISSING:
            self._ui.message_label.setText(_ERROR_MESSAGE.format("Domain name missing."))
        elif self._error_flags & MergingErrorFlag.NO_PARAMETER_SELECTED:
            self._ui.message_label.setText(_ERROR_MESSAGE.format("No source domain selected."))
        elif self._error_flags & MergingErrorFlag.DOMAIN_NAME_CLASH:
            self._ui.message_label.setText(_ERROR_MESSAGE.format("Domain name clashes with an existing set."))
        else:
            self._ui.message_label.setText("")

    def _clear_flag(self, state):
        """Clears a state flag."""
        if not self._error_flags:
            return
        self._error_flags &= ~state
        self._check_state()

    def _set_flag(self, state):
        """Sets a state flag."""
        self._error_flags |= state
        self._check_state()

    def _reset_indexing_domains_label(self, domain_name=None, domain_names=None):
        """Rewrites the contents of indexing_domains_label."""
        if domain_name is None:
            domain_name = self._ui.domain_name_edit.text()
        bold_name = "<b>{}</b>".format(domain_name if domain_name else "unnamed")
        if domain_names is None:
            if self._selected_domain_row is not None:
                domain_names = self._domain_names_model.item_at(self._selected_domain_row).domain_names
            else:
                domain_names = tuple()
        label = (
            "("
            + ", ".join(domain_names[: self._index_position] + (bold_name,) + domain_names[self._index_position :])
            + ")"
        )
        self._ui.indexing_domains_label.setText(label)

    @Slot(str)
    def _update_parameter_name(self, name):
        """Updates the merged parameter name."""
        self._parameter_name = name
        if not name:
            self._set_flag(MergingErrorFlag.PARAMETER_NAME_MISSING)
        else:
            self._clear_flag(MergingErrorFlag.PARAMETER_NAME_MISSING)

    @Slot(bool)
    def _remove_self(self, _):
        """Requests removal from the parent window."""
        self.removal_requested.emit(self)

    @Slot(QItemSelection, QItemSelection)
    def _handle_domain_selection_change(self, selected, _):
        """Resets the settings after another item has been selected in domains_list_view."""
        self._selected_domain_row = selected.indexes()[0].row()
        entity_class_info = self._domain_names_model.item_at(self._selected_domain_row)
        domain_names = entity_class_info.domain_names
        self._index_position = len(domain_names)
        self._parameter_name_list_model.reset_names(entity_class_info.parameter_names, entity_class_info.name)
        self._reset_indexing_domains_label(domain_names=domain_names)
        self._clear_flag(MergingErrorFlag.NO_PARAMETER_SELECTED)

    @Slot(str)
    def _update_indexing_domain_name(self, name):
        """Resets indexing_domains_label."""
        self._reset_indexing_domains_label(domain_name=name)
        if not name:
            self._set_flag(MergingErrorFlag.DOMAIN_NAME_MISSING)
        else:
            self._clear_flag(MergingErrorFlag.DOMAIN_NAME_MISSING)
            if name in self._set_settings.set_names:
                self._set_flag(MergingErrorFlag.DOMAIN_NAME_CLASH)
            else:
                self._clear_flag(MergingErrorFlag.DOMAIN_NAME_CLASH)

    @Slot(bool)
    def _move_domain_left(self, _):
        """Moves the new indexing domain left in indexing_domains_label."""
        if self._index_position > 0:
            self._index_position -= 1
            self._reset_indexing_domains_label()

    @Slot(bool)
    def _move_domain_right(self, _):
        """Moves the new indexing domain left in indexing_domains_label."""
        if self._selected_domain_row is None:
            return
        domain_names = self._domain_names_model.item_at(self._selected_domain_row).domain_names
        if domain_names and self._index_position < len(domain_names):
            self._index_position += 1
            self._reset_indexing_domains_label()
